import os
import math
import rasterio as rio
from movdata.base import CRS
from collections import UserDict


class RasterExtent:
    # minx, miny, maxx, maxy
    def __init__(self, x_min=33.0, x_max=37.0, y_min=2.0, y_max=-2.0):
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max

    def __repr__(self):
        return "RasterExtent(xmin: {0}, xmax: {1}, ymin: {2}, ymax: {3})".format(self.x_min,
                                                                                 self.x_max,
                                                                                 self.y_min,
                                                                                 self.y_max)

    @staticmethod
    def create_from_origin(pixel_size=0.5, x_min=0.0, y_min=0.0, num_rows=100, num_columns=100):
        max_x = x_min + (num_columns * pixel_size)
        max_y = y_min + (num_rows * pixel_size)
        return RasterExtent(x_min=x_min, x_max=max_x, y_min=y_min, y_max=max_y)


class RasterProfile(UserDict):
    """
    A class for holding raster properties
    At present this class is only valid for non-rotated rasters with a north-up orientation and square sized pixels
    defined by the E-W pixel size
    """

    def __init__(self,
                 pixel_size=1000.,
                 pixel_dtype=rio.float64,
                 crs='EPSG:8857',  # WGS 84/Equal Earth Greenwich
                 nodata_value=0.,
                 band_count=1,
                 raster_extent=None,
                 ):
        crs = CRS.from_user_input(crs)
        raster_extent = raster_extent or RasterExtent()
        cols = int(math.ceil((raster_extent.x_max - raster_extent.x_min) / pixel_size))
        rows = int(math.ceil((raster_extent.y_max - raster_extent.y_min) / pixel_size))
        affine_transform = rio.Affine(pixel_size, 0., raster_extent.x_min,
                                      0., -pixel_size, raster_extent.y_max)
        UserDict.__init__(self,
                          crs=crs,
                          pixel_size=pixel_size,
                          nodata=nodata_value,
                          columns=cols,
                          rows=rows,
                          dtype=pixel_dtype,
                          band_count=band_count,
                          raster_extent=raster_extent,
                          transform=affine_transform)

    def _recompute_transform_(self, key):
        """Recomputes the affine transformation matrix when the pixel_size or raster_extent value is updated."""
        _names = ['pixel_size', 'raster_extent']
        if all([key in _names, hasattr(self, _names[0]), hasattr(self, _names[1])]):
            xmin = self.raster_extent.x_min
            ymax = self.raster_extent.y_max
            pixel_size = self.pixel_size

            transform = rio.Affine(pixel_size, 0., xmin, 0., -pixel_size, ymax)
            cols = int(math.ceil((self.raster_extent.x_max - self.raster_extent.x_min) / self.pixel_size))
            rows = int(math.ceil((self.raster_extent.y_max - self.raster_extent.y_min) / self.pixel_size))

            self.update(transform=transform, columns=cols, rows=rows)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if name == 'data':
            object.__setattr__(self, name, value)
        else:
            self[name] = value
            self._recompute_transform_(name)

    def __setitem__(self, key, item):
        self.data[key] = item
        self._recompute_transform_(key)


class RasterPy:
    @classmethod
    def write(cls, ndarray, fp, columns, rows, band_count, driver='GTiff', dtype=rio.float64,
              crs=None, transform=None, nodata=None, sharing=False, indexes=1, **kwargs):
        with rio.open(fp=fp, mode='w', driver=driver,
                      height=rows, width=columns, count=band_count, dtype=dtype,
                      crs=crs, nodata=nodata,
                      sharing=sharing,
                      transform=transform) as r:
            r.write(ndarray, indexes=indexes)

    @classmethod
    def read(cls, fp, driver=None, **kwargs):
        fp = os.fspath(fp)
        path = rio.parse_path(fp)
        return rio.DatasetReader(path, driver=driver, **kwargs)

    @classmethod
    def read_write(cls, fp, driver=None, **kwargs):
        fp = os.fspath(fp)
        path = rio.parse_path(fp)
        return rio.get_writer_for_path(path, driver=driver)(path, 'r+', driver=driver, **kwargs)



from movdata import base
import datetime as dt
import math
import numpy as np
from osgeo import gdal
import tempfile
import os
import shutil
from joblib import Parallel, delayed
from joblib import load, dump
import movdata.raster as rst


class ETDAnalysisParams(base.AnalysisParams):

    def __init__(self,
                 maxdatagap_secs=14400,
                 maxspeed_kmhr=0.0,
                 maxspeed_percent=0.0,
                 intstep_kmhr=0.001,
                 etd_kernel=None,
                 prob_dens_cut_off=0.000001,
                 rast_params=None,
                 expansion_factor=1.3,
                 raster_storage=None):

        self._maxdatagap_secs = maxdatagap_secs
        self._maxspeed_kmhr = maxspeed_kmhr
        self._maxspeed_percent = maxspeed_percent
        self._intstep_kmhr = intstep_kmhr
        self._prob_dens_cut_off = prob_dens_cut_off
        self._etd_kernel = etd_kernel
        self._rast_params = rast_params
        self._expansion_factor = expansion_factor
        self._raster_storage = raster_storage


    @property
    def maxdatagap_secs(self):
        return self._maxdatagap_secs

    @property
    def maxspeed_kmhr(self):
        return self._maxspeed_kmhr

    @property
    def maxspeed_percent(self):
        return self._maxspeed_percent

    @property
    def intstep_kmhr(self):
        return self._intstep_kmhr

    @property
    def prob_dens_cut_off(self):
        return self._prob_dens_cut_off

    @property
    def rast_params(self):
        return self._rast_params

    @property
    def expansion_factor(self):
        return self._expansion_factor

    @property
    def raster_storage(self):
        return self._raster_storage

class ETDKernel:
    pass


class Weibull2Params(ETDKernel):

    def __init__(self,
                 shape=1.0,
                 scale=1.0):

        self._shape = shape
        self._scale = scale

    @property
    def scale(self):
        return self._scale

    @property
    def shape(self):
        return self._shape


class Weibull3Params(ETDKernel):
    """scale=a(t^b)(c^t) where t=time"""

    def __init__(self,
                 shape=1.0,
                 a=1.0,
                 b=1.0,
                 c=1.0):

        self._shape = shape
        self._a = a
        self._b = b
        self._c = c

    @property
    def scale(self):
        return self._scale

    @property
    def a(self):
        return self._a

    @property
    def b(self):
        return self._b

    @property
    def c(self):
        return self._c


class ETDAnalysisResult(base.AnalysisResult):

    def __init__(self, out_rast=None):
        self._out_rast = out_rast

    @property
    def out_rastdataset(self):
        return self._out_rast

    @out_rastdataset.setter
    def out_rastdataset(self, value):
        self._out_rast = value


class ETDAnalysis:

    @classmethod
    def calc_etd_range_parallel(cls, etd_analysis_params=None, trajectory=None):

        if __name__ != "__main__":
            # Create kde_analysis_result object
            etd_analysis_result = ETDAnalysisResult()
            etd_analysis_result.analysis_start = dt.datetime.utcnow()

            # Create a temporary folder
            folder = tempfile.mkdtemp()

            try:
                # Test to make sure the calculation parameters are not None and the correct type
                assert isinstance(etd_analysis_params, ETDAnalysisParams)

                # Make sure that the trajectory is not None
                if trajectory is None:
                    raise base.PymetException('KDE Analysis Error: Trajectory was None.')

                # Retrieve the Geopandas series for the trajectory
                traj_series = trajectory.relocs.as_geopandas_series()

                # Project to the desired output SRS
                if etd_analysis_params.rast_params.srs_code != 4326:
                    traj_series = traj_series.to_crs(epsg=etd_analysis_params.rast_params.srs_code)

                # Retrieve the x,y coordinates of the relocs into a numpy array
                traj_coords = np.array([[i[1].x, i[1].y] for i in traj_series.iteritems()])

                # Determine the envelope of the trajectory
                # env = trajectory.relocs().ogr_geometry().GetEnvelope()
                # x_min, x_max, y_min, y_max = env[0], env[1], env[2], env[3]
                (x_min, x_max, y_min, y_max) = trajectory.relocs.ogr_geometry.GetEnvelope()
                x_min = traj_coords[:, 0].min()
                x_max = traj_coords[:, 0].max()
                y_min = traj_coords[:, 1].min()
                y_max = traj_coords[:, 1].max()

                # Create a RasterProps object to define characteristics of the output raster
                raster_props = rst.RasterProps()

                # Determine the envelope wrt to the expansion_factor
                if etd_analysis_params.expansion_factor > 1.0:
                    dx = (x_max - x_min) * (etd_analysis_params.expansion_factor - 1.0) / 2.0
                    dy = (y_max - y_min) * (etd_analysis_params.expansion_factor - 1.0) / 2.0
                    x_min -= dx
                    x_max += dx
                    y_min -= dy
                    y_max += dy

                    # Determine the output raster size
                    num_columns = int(math.ceil((x_max - x_min) / etd_analysis_params.rast_params.pixel_size))
                    num_rows = int(math.ceil((y_max - y_min) / etd_analysis_params.rast_params.pixel_size))
                    # print('Columns & Rows:', num_columns, num_rows)

                    # Create the output raster file
                    driver = gdal.GetDriverByName('HFA')
                    outputpath = os.path.join(etd_analysis_params.raster_storage.location,
                                              etd_analysis_params.raster_storage.name)
                    output_raster = driver.Create(utf8_path=outputpath,
                                                  xsize=num_columns,
                                                  ysize=num_rows,
                                                  bands=1,
                                                  eType=gdal.GDT_Float64)

                    # Add the raster to the results object
                    etd_analysis_result.out_rastdataset = output_raster

                    # Define the affine transform for converting between grid coordinates and map coordinates
                    # needed by GDAL
                    transform_vals = (x_min, etd_analysis_params.rast_params.pixel_size, 0,
                                      y_max, 0, -etd_analysis_params.rast_params.pixel_size)
                    output_raster.SetGeoTransform(transform_vals)

                    # Get a hold of the raster output band
                    outband = output_raster.GetRasterBand(1)

                    # Set the NoData Value on the output band
                    outband.SetNoDataValue(etd_analysis_params.rast_params.no_data_value)

                    # Define the output_raster spatial reference
                    output_raster.SetProjection(base.SRS.srs_from_epsg
                                                (etd_analysis_params.rast_params.srs_code).ExportToWkt())

                    # Define the affine transform to get grid pixel centroids as geographic coordinates and adjusting
                    # the origin by a half pixel to the east and south from the grid origin point
                    img_centroids_to_map = np.array(transform_vals).reshape(2, 3)
                    img_centroids_to_map[0, 0] = x_min + etd_analysis_params.rast_params.pixel_size * 0.5
                    img_centroids_to_map[1, 0] = y_max - etd_analysis_params.rast_params.pixel_size * 0.5
                    # print(img_centroids_to_map)


            except base.PymetException as e:
                etd_analysis_result.add_error(e)

            finally:
                try:
                    shutil.rmtree(folder)
                except:
                    print("Failed to delete: " + folder)

            # Return the analysis kde_analysis_result
            etd_analysis_result.analysis_end = dt.datetime.utcnow()
            return etd_analysis_result  # return results


import movdata.base
import movdata.utils
import datetime as dt
from osgeo import ogr


class Geofence:
    """
    A polyline geometry representing a virtual fenceline
    ogr_geometry: can be anything but geofencing works on the intersection of a line with the input geoemtry
    """
    def __init__(self, ogr_geometry=None, fence_name='', unique_id='', warn_level=''):
        self._ogr_geometry = ogr_geometry
        self._fence_name = fence_name
        self._unique_id = unique_id
        self._warn_level = warn_level

    @property
    def name(self):
        return str(self._fence_name)

    @property
    def ogr_geometry(self):
        return self._ogr_geometry

    @property
    def unique_id(self):
        return self._unique_id

    @property
    def warn_level(self):
        return self._warn_level


class GeofenceAnalysisParams(movdata.base.AnalysisParams):
    """
    A class to hold parameters needed for the geofence calculation.

    geofences: a list of geofences to run against each trajectory
    geofenceregions: a list of regions to test containment before/after a detected geofence crossing
    """

    def __init__(self, geofences=None, regions=None):
        self._geofences = geofences or []
        self._regions = regions or []

    @property
    def geofences(self):
        return self._geofences

    @property
    def regions(self):
        return self._regions


class GeofenceAnalysisResult(movdata.base.AnalysisResult):
    """ A class to hold the results of a geofence analysis"""

    def __init__(self, geofence_crossings=None):
        self._geofence_crossings = geofence_crossings or []

    @property
    def geofence_crossings(self):
        return self._geofence_crossings

    def add_crossing(self, geofence_crossing=None):
        if geofence_crossing is not None:
            self._geofence_crossings.append(geofence_crossing)


class GeofenceCrossing:

    """ Class to store the result of a single geofence crossing"""

    def __init__(self, subject_id='', subject_speed=0.0, subject_travel_heading=0.0,
                 estimated_cross_fix=None, geofence_id='', warn_level='',
                 start_region_ids=None, end_region_ids=None):
        self._subject_id = subject_id
        self._est_cross_fix = estimated_cross_fix
        self._start_region_ids = start_region_ids or []
        self._end_region_ids = end_region_ids or []
        self._geofence_id = geofence_id
        self._warn_level = warn_level
        self._subject_speed_kmhr = subject_speed
        self._subject_heading = subject_travel_heading

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def subject_speed_kmhr(self):
        return self._subject_speed_kmhr

    @property
    def subject_heading(self):
        return self._subject_heading

    @property
    def est_cross_fix(self):
        return self._est_cross_fix

    @property
    def start_region_ids(self):
        return self._start_region_ids

    @property
    def end_region_ids(self):
        return self._end_region_ids

    @property
    def geofence_id(self):
        return self._geofence_id

    @property
    def warn_level(self):
        return self._warn_level


class GeofenceAnalysis:

    @classmethod
    def calc_crossings(cls, geofence_analysis_params=None, trajectories=None):
        """
        Run the crossings analysis using the input fences/regions against the various
        :param geofence_analysis_params:
        :param trajectories:
        :return:
        """

        trajectories = trajectories or []

        # Test to make sure the calculation parameters are not None and the correct type
        assert type(geofence_analysis_params) is GeofenceAnalysisParams

        # Create the output analysis result object
        result = GeofenceAnalysisResult()

        # Set the start time of the analysis
        result.analysis_start = dt.datetime.utcnow()

        for traj in trajectories:
            assert type(traj) is movdata.base.Trajectory

            subject_id = traj.relocs.subject_id
            trajsegs = traj.traj_segs

            for trajseg in trajsegs:
                fences = geofence_analysis_params.geofences
                for fence in fences:
                    assert type(fence) is Geofence

                    # Attempt the intersection of the trajectory segment with the fence
                    intersect_pnts = trajseg.ogr_geometry.Intersection(fence.ogr_geometry)

                    # intersect_pnts can either be None, POINT, or MULTIPOINT
                    _intersectPnts = []

                    # TODO better way to asses ogr geometry type?
                    if intersect_pnts.GetGeometryName() == 'POINT':
                        _intersectPnts.append(intersect_pnts)
                    else:
                        _intersectPnts = intersect_pnts

                    for pnt in _intersectPnts:

                        # Create a GeoPoint at the crossing OGR point
                        crossing_geopoint = movdata.base.GeoPoint(x=pnt.GetX(), y=pnt.GetY())

                        segment_distance_to_crossing = \
                            trajseg.start_fix_geopoint.dist_to_point(pnt)

                        segment_length = trajseg.length_km * 1000.0

                        fractional_distance = 0.0
                        if segment_length > 0.0:
                            fractional_distance = segment_distance_to_crossing / segment_length

                        # Estimate the time of the break based on the fractional timespan
                        fractional_time = fractional_distance * (trajseg.end_fix.fixtime - trajseg.start_fix.fixtime)
                        crossing_time = trajseg.start_fix.fixtime + fractional_time

                        # Create an estimated geofence cross fix
                        estimated_cross_fix = movdata.base.Fix(crossing_geopoint, crossing_time)

                        # Determine containment of the subject before and after the crossing
                        containment_before = GeofenceAnalysis.asses_containment(trajseg.start_fix_geopoint,
                                                                                geofence_analysis_params.regions)

                        containment_after = GeofenceAnalysis.asses_containment(trajseg.end_fix_geopoint,
                                                                               geofence_analysis_params.regions)

                        # Create the output fence crossing result
                        crossing = GeofenceCrossing(subject_id=subject_id,
                                                    subject_speed=trajseg.speed_kmhr,
                                                    subject_travel_heading=trajseg.heading,
                                                    estimated_cross_fix=estimated_cross_fix,
                                                    geofence_id=fence.unique_id,
                                                    warn_level=fence.warn_level,
                                                    start_region_ids=containment_before,
                                                    end_region_ids=containment_after)

                        # Add this given crossing to the result
                        result.add_crossing(crossing)

        # TODO: Sort the crossings chronologically

        # Set the end time of the analysis
        result.analysis_end = dt.datetime.utcnow()

        return result

    @classmethod
    def equator_fence(cls):
        """ Create a virtual fence around the equator that can be used for testing"""

        equatorFenceSeg = ogr.Geometry(type=ogr.wkbLineString)
        equatorFenceSeg.AssignSpatialReference(movdata.base.SRS.WGS84())
        equatorFenceSeg.AddPoint(0, 0, 0)
        equatorFenceSeg.AddPoint(90, 0, 0)
        equatorFenceSeg.AddPoint(180, 0, 0)
        equatorFenceSeg.AddPoint(270, 0, 0)
        equatorFenceSeg.AddPoint(0, 0, 0)

        #Wrap the Linestring into a Multilinestring for kicks
        equatorFence = ogr.Geometry(type=ogr.wkbMultiLineString)
        equatorFence.AssignSpatialReference(movdata.base.SRS.WGS84())
        equatorFence.AddGeometry(equatorFenceSeg)

        return equatorFence

    @classmethod
    def asses_containment(cls, geopoint=None, regions=None):
        return movdata.utils.assess_containment(geopoint=geopoint, regions=regions)


import datetime as dt
import functools
import uuid as guid
import pytz
import typing

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import osgeo
import shapely.geometry as shp
import shapely.ops as ops
import shapely.wkt as wkt
from pyproj import Geod
from pyproj.enums import WktVersion
from osgeo import ogr
from osgeo import osr
from pyarrow import feather, parquet
import pygeos
from dataclasses import dataclass, field


class __WGS84Meta(type):
    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(mcs, name, bases, attrs)
        new_class.WGS84 = pyproj.crs.CRS.from_epsg(4326)
        return new_class


class CRS(metaclass=__WGS84Meta):

    @classmethod
    def crs_from_epsg(cls, epsg=4326):
        return pyproj.crs.CRS.from_epsg(epsg)

    @classmethod
    def crs_from_wkt(cls, wkt_string):
        return pyproj.crs.CRS.from_wkt(wkt_string)

    @classmethod
    def crs_from_wkt_str(cls, txt='ESPG:4326'):
        return pyproj.crs.CRS.from_string(txt)

    @classmethod
    def epsg_from_crs(cls, crs):
        return crs.to_espg()

    @classmethod
    def wkt_from_crs(cls, crs):
        return crs.to_wkt()

    @classmethod
    def from_user_input(cls, crs):
        return pyproj.CRS.from_user_input(crs)

    @classmethod
    def crs_from_srs(cls, srs):
        """convert osgeo.osr.SpatialReference to pyproj.crs.CRS"""
        if osgeo.version_info.major < 3:
            return pyproj.crs.CRS.from_wkt(srs.ExportToWkt())
        else:
            return pyproj.crs.CRS.from_wkt(srs.ExportToWkt(["FORMAT=WKT2_2018"]))

    @classmethod
    def srs_from_crs(cls, crs):
        """Convert from pyproj.crs.CRS to osgeo.osr.SpatialReference

        WKT2 is the best format for storing your CRS according to the PROJ FAQ and
        only supported in GDAL 3+
        """
        crs = cls.from_user_input(crs)

        srs = osr.SpatialReference()
        if osgeo.version_info.major < 3:
            srs.ImportFromWkt(crs.to_wkt(WktVersion.WKT1_GDAL))
        else:
            srs.ImportFromWkt(crs.to_wkt())
        return srs


def pyproj_transformer(crs, crs2):
    _crs = pyproj.CRS.from_user_input(crs)
    to_crs = pyproj.CRS.from_user_input(crs2)
    return pyproj.Transformer.from_crs(_crs, to_crs, always_xy=True)


@dataclass
class RelocsCoordinateFilter:
    """Filter parameters for filtering get_fixes based on X/Y coordinate ranges or specific coordinate values"""
    min_x: float = -180.
    max_x: float = 180.
    min_y: float = -90.
    max_y: float = 90.
    filter_point_coords: typing.Union[typing.List[typing.List[float]], gpd.GeoSeries] = field(
        default_factory=[[0., 0.]])

    def __post_init__(self):
        if isinstance(self.filter_point_coords, list):
            self.filter_point_coords = gpd.GeoSeries(shp.Point(coord) for coord in self.filter_point_coords)


@dataclass
class RelocsDateRangeFilter:
    """Filter parameters for filtering based on a datetime range"""
    start: dt.datetime
    end: dt.datetime


@dataclass
class RelocsSpeedFilter:
    """Filter parameters for filtering based on the speed needed to move from one fix to the next"""
    max_speed_kmhr: float = float("inf")
    temporal_order: str = 'ASC'


@dataclass
class RelocsDistFilter:
    """Filter parameters for filtering based on the distance needed to move from one fix to the next"""

    max_dist_km: float = float("inf")
    temporal_order: str = "ASC"


@dataclass
class TrajSegFilter:
    """Class filtering a set of trajectory segment segments"""

    min_length_meters: float = 0.0
    max_length_meters: float = float("inf")
    min_time_secs: float = 0.0
    max_time_secs: float = float("inf")
    min_speed_kmhr: float = 0.0
    max_speed_kmhr: float = float("inf")


class BaseMixin:
    """
    A class that contains methods for use by other classes without having to be
    the parent class of those other classes.
    '"""
    @classmethod
    def get_value(cls, item: typing.Dict, keyname: str):
        """Get the value from the dictionary"""
        return item[keyname]

    @classmethod
    def set_metadata(cls, dataframe: typing.Union[gpd.GeoDataFrame, pd.DataFrame], columns: typing.List):
        """This method would add all extra columns to single column"""
        metadata_cols = set(dataframe.columns).difference(set(columns))

        # To prevent additional column from being dropped, name the column metadata (rename it back).
        dataframe['metadata'] = dataframe[metadata_cols].to_dict(orient='records')
        dataframe.drop(metadata_cols, inplace=True, axis=1)
        dataframe.rename(columns={'metadata': 'additional'}, inplace=True)
        return dataframe


class GeoPoint:
    """A geographic point is the most basic geometry"""

    def __new__(cls, _pnt=None, *args, **kwargs):
        instance = object.__new__(cls)
        instance._shapely_geo = _pnt
        return instance

    def __init__(self, x: float = 0., y: float = 0., z: float = 0., crs=CRS.WGS84):
        """Create a new point with the supplied geographic coordinates.
        Defaults to WGS84"""
        self._point = self._shapely_geo or shp.Point(x, y, z)
        self.crs = crs

    @property
    def shapely_geometry(self):
        return self._point

    @property
    def X(self):
        """Shortcut to getting the X value"""
        return self._point.x

    @property
    def Y(self):
        """Shortcut to getting the Y value"""
        return self._point.y

    @property
    def wkt_geometry(self):
        return self._point.wkt

    @property
    def ogr_geometry(self):
        """ Return the ogr point """
        return ogr.CreateGeometryFromWkt(self.wkt_geometry)

    def to_crs(self, crs):
        """https://shapely.readthedocs.io/en/stable/manual.html#shapely.ops.transform"""
        if isinstance(crs, pyproj.transformer.Transformer):
            return ops.transform(crs.transform, self.shapely_geometry)
        tranformer = pyproj_transformer(self.crs, crs)
        return ops.transform(tranformer.transform, self.shapely_geometry)

    def __str__(self):
        return str(self._point)


class Fix:
    """A Fix is a temporal geopoint"""

    def __init__(self, geopoint=None,
                 fixtime=None,
                 filter_status=False,
                 fix_id=None,
                 additional_data=None):
        self._fixtime = fixtime
        self._geopoint = geopoint
        self._filter_status = filter_status  # Holds a boolean value to indicate whether this point is junk or not
        self._fix_id = fix_id or str(guid.uuid4().hex)
        self._additional_data = additional_data or dict()

    @property
    def fixtime(self):
        """ Return the fixtime"""
        return self._fixtime

    @property
    def geopoint(self):
        """Return the geopoint"""
        return self._geopoint

    # Eventually might want to support geoson_geometry

    @property
    def shapely_geometry(self):
        if self._geopoint is not None:
            return self._geopoint.shapely_geometry

    @property
    def wkt_geometry(self):
        return self.shapely_geometry.wkt

    @property
    def ogr_geometry(self):
        """A shortcut for returning the ogr point geometry from the geopoint"""
        if self.shapely_geometry is not None:
            return ogr.CreateGeometryFromWkt(self.wkt_geometry)

    @property
    def filter_status(self):
        return self._filter_status

    @filter_status.setter
    def filter_status(self, value):
        assert type(value) is bool
        self._filter_status = value

    @property
    def fix_id(self):
        return self._fix_id

    @property
    def additional_data(self):
        return self._additional_data

    @staticmethod
    def from_dict(in_dict=None,
                  id_key='observation_id',
                  fixtime_key='fixtime',
                  lat_key='lat',
                  lon_key='lon',
                  height_key='height',
                  junk_key=None):

        """ A function to create a Fix object from a dictionary object"""

        if isinstance(in_dict, dict):

            # Create a copy
            tmp_dict = in_dict.copy()  # is this too expensive?

            id_val = tmp_dict.get(id_key, None)
            fixtime_val = tmp_dict.get(fixtime_key, None)
            lat_val = tmp_dict.get(lat_key, 0.)
            lon_val = tmp_dict.get(lon_key, 0.)
            height_val = tmp_dict.get(height_key, 0.)
            junk_val = tmp_dict.get(junk_key, False)

            # check that the id value is either a string or UUID object
            if not isinstance(id_val, (guid.UUID, str)):
                raise Exception('Create fix from dict error: the supplied uuid value is not a string')

            # check that the fixtime val is a dt.datetime  (will throw an error if can't be cast
            fixtime_val = pd.to_datetime(fixtime_val, errors='raise')

            # check that the lat_val is a floating type
            if not isinstance(lat_val, (np.floating, float)):
                raise Exception('Create fix from dict error: the supplied lat value is not a float')

            # check that the lat_val is a floating type
            if not isinstance(lon_val, (np.floating, float)):
                raise Exception('Create fix from dict error: the supplied lon value is not a float')

            # check that the lat_val is a floating type
            if not isinstance(height_val, (np.floating, float)):
                raise Exception('Create fix from dict error: the supplied height value is not a float')

            # check that the junk flag is a boolean type
            if not isinstance(junk_val, bool):
                raise Exception('Create fix from dict error: the supplied junk flag value is not a boolean')

            # drop the first class keys from the dict so that everything else is additional_data
            tmp_dict.pop(id_key, None)
            tmp_dict.pop(fixtime_key, None)
            tmp_dict.pop(lat_key, None)
            tmp_dict.pop(lon_key, None)
            tmp_dict.pop(height_key, None)
            tmp_dict.pop(junk_key, None)

            return Fix(GeoPoint(lon_val, lat_val, height_val),
                       fixtime=fixtime_val,
                       filter_status=junk_val,
                       fix_id=id_val,
                       additional_data=tmp_dict)

        else:
            return None


def straighttrack_properties(df: gpd.GeoDataFrame):
    """Make sure dataframe gets explicitly converted to WGS84 projection."""

    class Properties:
        @property
        def start_fixes(self):
            # unpack xy-coordinates of start fixes
            return df['geometry'].x, df['geometry'].y

        @property
        def end_fixes(self):
            # unpack xy-coordinates of end fixes
            return df['_geometry'].x, df['_geometry'].y

        @property
        def inverse_transformation(self):
            # use pyproj geodesic inverse function to compute vectorized distance & heading calculations
            return Geod(ellps="WGS84").inv(*self.start_fixes, *self.end_fixes)

        @property
        def heading(self):
            # Forward azimuth(s)
            forward_azimuth, _, _ = self.inverse_transformation
            forward_azimuth[forward_azimuth < 0] += 360
            return forward_azimuth

        @property
        def dist_meters(self):
            _, _, distance = self.inverse_transformation
            return distance

        @property
        def timespan_seconds(self):
            return (df['_fixtime'] - df['fixtime']).dt.total_seconds()

        @property
        def speed_kmhr(self):
            return (self.dist_meters / self.timespan_seconds) * 3.6

        @property
        def x_displacement(self):
            return self.end_fixes[0] - self.start_fixes[0]

        @property
        def y_displacement(self):
            return self.end_fixes[1] - self.start_fixes[1]

    instance = Properties()
    return instance


class Relocations:
    """Relocations is a model for a set of fixes from a given subject.
    Because fixes are temporal, they can be ordered asc or desc. The additional_data dict can contain info
    specific to the subject and relocations: name, type, region, sex etc. These values are applicable to all
    fixes in the relocations array. If they vary, then they should be put into each fix's additional_data dict.
    """

    # TODO: Net Square Displacement
    # TODO: Correlated Random Walk from start fix
    # TODO: Biased Random Walk from start fix
    # TODO: Centroid as the harmonic mean of coordinates (see Avagar et al. 2012)

    def __init__(self, fixes=None, additional_data=None, crs=CRS.WGS84):
        self._fixes = fixes
        self._crs = crs
        self._gdf = self._create_dataframe(fixes)
        self._additional_data = additional_data or dict()

    def _create_dataframe(self, fixes):
        """Create Relocations dataframe. Get all the tranformer instances needed for different reprojection.
        """
        if isinstance(fixes, gpd.GeoDataFrame):
            return fixes.loc[:, ['subject_id', 'fixtime', 'geometry', 'filter_status', 'additional']]
        elif fixes is not None:
            self._crs_strings = set(np.vectorize(lambda x: str(x.geopoint.crs))(self._fixes))
            self.transformers = {crs_str: pyproj_transformer(crs_str, self._crs) for crs_str in self._crs_strings}
            vfunc = np.vectorize(self._fix)
            return gpd.GeoDataFrame(vfunc(fixes).tolist(), crs=self._crs)
        else:
            return gpd.GeoDataFrame()

    def _fix(self, f):
        tranformer = self.transformers.get(str(f.geopoint.crs))
        return pd.Series({'uuid': f.fix_id,
                          'fixtime': f.fixtime,
                          'geometry': f.geopoint.to_crs(tranformer),
                          'filter_status': f.filter_status,
                          'additional': f.additional_data
                          })

    def df(self, temporal_order='ASC', filtered=False):
        ascending = True if temporal_order == 'ASC' else False
        return self._gdf[self._gdf.filter_status == filtered].sort_values(by='fixtime', ascending=ascending)

    _dataframe = property(df)

    @property
    def timespan_seconds(self):
        """Returns the total time between start and end times in seconds"""
        if not self._dataframe.empty:
            earliest_fixtime = self.earliest_fix.fixtime
            latest_fixtime = self.latest_fix.fixtime

            if (earliest_fixtime is not None) and (latest_fixtime is not None):
                ts = latest_fixtime - earliest_fixtime
                return ts.total_seconds()
        else:
            return 0

    @property
    def subject_id(self):
        if not self._gdf.empty:
            return self._gdf.subject_id.unique()
        return

    @property
    def additional_data(self):
        return self._additional_data

    @property
    def fix_count(self):
        """ get_fixes array length"""
        return self._dataframe.shape[0]

    @property
    def earliest_fix(self):
        """ Return the earliest fix from the get_fixes array"""
        _df = self._dataframe
        if not _df.empty:
            return _df.iloc[0]
        else:
            return None

    @property
    def latest_fix(self):
        """ Return the latest fix from the get_fixes array"""
        _df = self._dataframe
        if not _df.empty:
            return _df.iloc[-1]
        else:
            return None

    @property
    def shapely_geometry(self):
        """Return the shapely multipoint geometry of the relocation get_fixes"""
        _pnts = shp.MultiPoint(self._dataframe.geometry)
        return _pnts

    @property
    def wkt_geometry(self):
        return self.shapely_geometry.wkt

    @property
    def ogr_geometry(self):
        """Return the ogr multipoint geometry of the relocation get_fixes"""
        return ogr.CreateGeometryFromWkt(self.wkt_geometry)

    def clip(self, region=None):

        if not type(region) is Region:
            return None

        if self._gdf.empty:
            return None

        if self._gdf.crs == region.crs:
            self._gdf = gpd.clip(self._gdf, Region.shapely_geometry)

        return self

    def to_crs(self, crs):  # rename to_crs
        """function to change coordinate reference system."""
        return self._dataframe.to_crs(crs)

    @property
    def centroid(self):
        """returns GeoSeries of centroids"""
        return self._dataframe.centroid

    def denormalize_metadata(self, keyname: typing.Union[str, typing.List[str]]):
        """function to extract values from metadata (additional) column and create independent column"""
        if isinstance(keyname, str):
            self._gdf[keyname] = self._gdf['additional'].map(functools.partial(BaseMixin.get_value, keyname=keyname))
        elif isinstance(keyname, list):
            for key in keyname:
                self._gdf[key] = self._gdf['additional'].map(functools.partial(BaseMixin.get_value, keyname=key))

    @staticmethod
    def _to_geopandas(df,
                      subject_field,
                      x_field,
                      y_field,
                      wkt_field,
                      wkb_field,
                      fixtime_field,
                      status_field,
                      timezone,
                      crs):
        df[fixtime_field] = pd.to_datetime(df[fixtime_field], utc=timezone)

        if not status_field:
            df[status_field] = False

        if isinstance(df, gpd.GeoDataFrame):
            # make a copy of geodataframe to avoid pointing two variables to the same object in memory.
            gdf = df.copy()
        else:
            if wkt_field:
                geometry = df[wkt_field].apply(lambda x: wkt.loads(x))
            elif wkb_field:
                geometry = gpd.array.from_wkb(df[wkb_field], crs=crs)
            elif x_field and y_field:
                geometry = gpd.points_from_xy(df[x_field], df[y_field])
            else:
                raise ValueError("""No geometry columns fields were passed in the parameters""")
            gdf = gpd.GeoDataFrame(df, geometry=geometry, crs=crs)

        columns = {subject_field: "subject_id",
                   fixtime_field: "fixtime",
                   status_field: "filter_status",
                   'geometry': 'geometry'}

        gdf.rename(columns=columns, inplace=True)
        gdf.index.rename('uuid', inplace=True)
        gdf = BaseMixin.set_metadata(gdf, list(columns.values()))
        return gdf

    @classmethod
    def from_csv(cls,
                 csv_file,
                 subject_field,
                 x_field='lon',
                 y_field='lat',
                 unique_field=None,
                 wkt_field=None,
                 wkb_field=None,
                 fixtime_field='fixtime',
                 filter_status_field=None,
                 timezone=pytz.UTC,
                 crs=CRS.WGS84,
                 header_row=0,
                 compression="infer"
                 ):

        df = pd.read_csv(csv_file, header=header_row, compression=compression)

        if unique_field is not None:
            df.set_index(unique_field, inplace=True)
        else:
            df.index = np.vectorize(lambda x: guid.uuid4())(range(df.shape[0]))

        gdf = cls._to_geopandas(df, subject_field=subject_field,
                                x_field=x_field, y_field=y_field,
                                wkt_field=wkt_field,
                                wkb_field=wkb_field,
                                fixtime_field=fixtime_field,
                                status_field=filter_status_field,
                                timezone=timezone,
                                crs=crs
                                )
        return cls(gdf, crs=crs)

    @classmethod
    def from_geodataframe(cls,
                          gdf,
                          fixtime_field,
                          subject_field='subject_id',
                          unique_field=None,
                          filter_status_field=None,
                          crs=CRS.WGS84,
                          timezone=pytz.UTC
                          ):

        if not isinstance(gdf, gpd.GeoDataFrame):
            raise TypeError('Expecting GeoDataFrame')

        if unique_field is not None:
            gdf.set_index(unique_field, inplace=True)
        else:
            gdf.index = np.vectorize(lambda x: guid.uuid4())(range(gdf.shape[0]))

        gdf = cls._to_geopandas(gdf, subject_field=subject_field,
                                x_field=None,
                                y_field=None,
                                wkt_field=None,
                                wkb_field=None,
                                fixtime_field=fixtime_field,
                                status_field=filter_status_field,
                                crs=crs,
                                timezone=timezone
                                )
        return cls(gdf, crs=crs)

    @classmethod
    def from_feather(cls, path,
                     subject_field,
                     x_field=None,
                     y_field=None,
                     wkt_field=None,
                     wkb_field='geometry',
                     fixtime_field='fixtime',
                     filter_status_field=None,
                     crs=CRS.WGS84,
                     timezone=pytz.UTC,
                     **kwargs):
        """Load a Feather object from the file path, returning a GeoDataFrame."""
        df = feather.read_table(path, **kwargs).to_pandas()

        gdf = cls._to_geopandas(df, subject_field=subject_field,
                                x_field=x_field, y_field=y_field,
                                wkt_field=wkt_field,
                                wkb_field=wkb_field,
                                fixtime_field=fixtime_field,
                                status_field=filter_status_field,
                                timezone=timezone,
                                crs=crs
                                )

        return cls(gdf, crs=crs)

    @classmethod
    def from_parquet(cls, path,
                     subject_field,
                     x_field=None,
                     y_field=None,
                     wkt_field=None,
                     wkb_field='geometry',
                     fixtime_field='fixtime',
                     filter_status_field=None,
                     crs=CRS.WGS84,
                     timezone=pytz.UTC,
                     **kwargs
                     ):
        kwargs["use_pandas_metadata"] = True
        df = parquet.read_table(path, **kwargs).to_pandas()

        gdf = cls._to_geopandas(df, subject_field=subject_field,
                                x_field=x_field, y_field=y_field,
                                wkt_field=wkt_field,
                                wkb_field=wkb_field,
                                fixtime_field=fixtime_field,
                                status_field=filter_status_field,
                                timezone=timezone,
                                crs=crs
                                )
        return cls(gdf, crs=crs)

    @classmethod
    def from_observations(cls, obs_df):
        if not obs_df.empty:
            df = obs_df.assign(filter_status=False)
            df.index.rename('uuid', inplace=True)
            df = BaseMixin.set_metadata(df, ['subject_id', 'fixtime', 'geometry', 'filter_status'])
        else:
            df = None
        return cls(df)

    def sort_fixes(self, temporal_order='ASC'):
        return self._gdf.sort_values(by='fixtime', ascending=True if temporal_order == 'ASC' else False)

    @staticmethod
    def _apply_speedfilter(df, fix_filter):
        gdf = df.assign(_fixtime=df['fixtime'].shift(-1),
                        _geometry=df['geometry'].shift(-1),
                        _filter_status=df['filter_status'].shift(-1))[:-1]

        straight_track = straighttrack_properties(gdf)
        gdf['speed_kmhr'] = straight_track.speed_kmhr

        gdf.loc[(gdf['filter_status'] == False) &
                (gdf['_filter_status'] == False) &
                (gdf['speed_kmhr'] > fix_filter.max_speed_kmhr), 'filter_status'] = True

        gdf.drop(['_fixtime', '_geometry', '_filter_status', 'speed_kmhr'], axis=1, inplace=True)
        return gdf

    @staticmethod
    def _apply_distfilter(df, fix_filter):
        gdf = df.assign(_filter_status=df['_filter_status'].shift(-1),
                        _geometry=df['geometry'].shift(-1))[:-1]

        _, _, distance_m = Geod(ellps="WGS84").inv(gdf['geometry'].x,
                                                   gdf['geometry'].y,
                                                   gdf['_geometry'].x,
                                                   gdf['_geometry'].y)
        gdf['distance_km'] = distance_m / 1000

        gdf.loc[(gdf['filter_status'] == False) &
                (gdf['_filter_status'] == False) &
                (gdf['distance_km'] > fix_filter.max_dist_km), 'filter_status'] = True

        gdf.drop(['_geometry', '_filter_status', 'distance_km'], axis=1, inplace=True)
        return gdf

    def apply_fix_filter(self, fix_filter=None):
        """Apply a given filter by marking the fix filter_status based on the conditions of a filter"""

        # Identify junk fixes based on location coordinate x,y ranges or that match specific coordinates
        if isinstance(fix_filter, RelocsCoordinateFilter):
            self._gdf.loc[(self._gdf['geometry'].x < fix_filter.min_x) |
                          (self._gdf['geometry'].x > fix_filter.max_x) |
                          (self._gdf['geometry'].y < fix_filter.min_y) |
                          (self._gdf['geometry'].y > fix_filter.max_y) |
                          (self._gdf['geometry'].isin(fix_filter.filter_point_coords)), 'filter_status'] = True

        # Mark fixes outside this date range as junk
        if isinstance(fix_filter, RelocsDateRangeFilter):
            if fix_filter.start is not None:
                self._gdf.loc[self._gdf['fixtime'] < fix_filter.start, 'filter_status'] = True

            if fix_filter.end is not None:
                self._gdf.loc[self._gdf['fixtime'] > fix_filter.end, 'filter_status'] = True

        # Mark fixes junk if object needed to move faster than the max_speed_kmhr value to reach next point
        if isinstance(fix_filter, RelocsSpeedFilter):
            self.sort_fixes(temporal_order='ASC')

            if self._gdf.crs == CRS.WGS84:
                self._gdf = self._apply_speedfilter(self._gdf, fix_filter)
            else:
                crs = self._gdf.crs
                self._gdf = self._gdf.to_crs(CRS.WGS84)
                self._gdf = self._apply_speedfilter(self._gdf, fix_filter)
                self._gdf = self._gdf.to_crs(crs)

                # Mark fixes junk if object needed to move greater than the max_dist_meters value to reach next point
        if isinstance(fix_filter, RelocsDistFilter):
            self.sort_fixes(temporal_order='ASC')

            if self._gdf.crs == CRS.WGS84:
                self._gdf = self._apply_distfilter(self._gdf, fix_filter)
            else:
                crs = self._gdf.crs
                self._gdf = self._gdf.to_crs(CRS.WGS84)
                self._gdf = self._apply_distfilter(self._gdf, fix_filter)
                self._gdf = self._gdf.to_crs(crs)

    def reset_fix_filter(self):
        self._gdf['filter_status'] = False


class MultiRelocations(Relocations):
    def apply_fix_filter(self, fix_filter=None):
        if isinstance(fix_filter, RelocsCoordinateFilter) or isinstance(fix_filter, RelocsDateRangeFilter):
            super().apply_fix_filter(fix_filter)

        if isinstance(fix_filter, RelocsSpeedFilter):
            if self._gdf.crs == CRS.WGS84:
                self._gdf = self._gdf.groupby('subject_id').apply(self._apply_speedfilter,
                                                                  fix_filter=fix_filter).droplevel(['subject_id'])
            else:
                crs = self._gdf.crs
                self._gdf = self._gdf.to_crs(CRS.WGS84)
                self._gdf = self._gdf.groupby('subject_id').apply(self._apply_speedfilter,
                                                                  fix_filter=fix_filter).droplevel(['subject_id'])
                self._gdf = self._gdf.to_crs(crs)

        if isinstance(fix_filter, RelocsDistFilter):
            if self._gdf.crs == CRS.WGS84:
                self._gdf = self._gdf.groupby('subject_id').apply(self._apply_distfilter,
                                                                  fix_filter=fix_filter,
                                                                  ).droplevel(['subject_id'])
            else:
                crs = self._gdf.crs
                self._gdf = self._gdf.to_crs(CRS.WGS84)
                self._gdf = self._gdf.groupby('subject_id').apply(self._apply_distfilter,
                                                                  fix_filter=fix_filter,
                                                                  ).droplevel(['subject_id'])

                self._gdf = self._gdf.to_crs(crs)


class Trajectory:
    """
    A trajectory represents a time-ordered collection of segments.
    Currently only straight track segments exist.
    It is based on an underlying relocs object that is the point representation
    """

    def __init__(self, relocs):  # todo: don't need crs param here - get from relocs object
        _relocs_df = relocs.df()
        _relocs_df['_fixtime'] = _relocs_df['fixtime'].shift(-1)
        _relocs_df['_geometry'] = _relocs_df['geometry'].shift(-1)
        self.crs = _relocs_df.crs

        if _relocs_df.crs == CRS.WGS84:
            self._gdf = self._create_trajsegments(_relocs_df[:-1])
        else:
            _relocs_df = _relocs_df.to_crs(CRS.WGS84)  # explicitly project to WGS84 for creating traj segments
            self._gdf = self._create_trajsegments(_relocs_df[:-1])
            self._gdf = self._gdf.to_crs(self.crs)  # reproject back to original crs if different from WGS84

        self.relocs_df = relocs.df()

    @staticmethod
    def _create_trajsegments(df):
        # TODO: future implementation: how can ew speed this up using a C dll?

        track_properties = straighttrack_properties(df)

        # pygeos.Geometry [LineStrings]
        coords = np.column_stack((np.column_stack(track_properties.start_fixes),
                                  np.column_stack(track_properties.end_fixes))).reshape(df.shape[0], 2, 2)
        geometry = pygeos.linestrings(coords)

        # assemble results in new dataframe
        return gpd.GeoDataFrame(
            {
                'subject_id': df.subject_id,
                'segment_start': df.fixtime,
                'segment_end': df._fixtime,
                'timespan_seconds': track_properties.timespan_seconds,
                'dist_meters': track_properties.dist_meters,
                'speed_kmhr': track_properties.speed_kmhr,
                'heading': track_properties.heading,
                'x_displacement': track_properties.x_displacement,
                'y_displacement': track_properties.y_displacement,
                'geometry': geometry,
                'filter_status': df.filter_status,
                'additional': df.additional,
            },
            crs=4326,
            index=df.index)

    def df(self, temporal_order='ASC', filtered=False):
        ascending = True if temporal_order == 'ASC' else False
        return self._gdf[self._gdf.filter_status == filtered].sort_values(by='segment_start', ascending=ascending)

    _dataframe = property(df)

    def denormalize_metadata(self, keyname: typing.Union[str, typing.List[str]]):
        """function to extract values from metadata (additional) column and create independent column"""
        if isinstance(keyname, str):
            self._gdf[keyname] = self._gdf['additional'].map(functools.partial(BaseMixin.get_value, keyname=keyname))
        elif isinstance(keyname, list):
            for key in keyname:
                self._gdf[key] = self._gdf['additional'].map(functools.partial(BaseMixin.get_value, keyname=key))

    @property
    def relocs(self):
        """Return the relocs object that make up the trajectory"""
        return self.relocs_df

    def apply_traj_filter(self, traj_seg_filter):
        assert type(traj_seg_filter) is TrajSegFilter
        self._gdf.loc[
            (self._gdf['dist_meters'] < traj_seg_filter.min_length_meters) |
            (self._gdf['dist_meters'] > traj_seg_filter.max_length_meters) |
            (self._gdf['timespan_seconds'] < traj_seg_filter.min_time_secs) |
            (self._gdf['timespan_seconds'] > traj_seg_filter.max_time_secs) |
            (self._gdf['speed_kmhr'] < traj_seg_filter.min_speed_kmhr) |
            (self._gdf['speed_kmhr'] > traj_seg_filter.max_speed_kmhr), 'filter_status'] = True

    @property
    def traj_seg_counts(self):
        """A shortcut to getting the length of the trajectory segments array"""
        return len(self._dataframe.index)

    @property
    def tortuosity(self):
        # TODO Need to implement this property
        return 1.0

    @property
    def length_km(self):
        return sum(self._dataframe['dist_meters'].to_list()) * 1000

    @property
    def shapely_geometry(self):
        """Create MultiLineString object"""
        return shp.MultiLineString(self._dataframe['geometry'].to_list())

    @property
    def ogr_geometry(self):
        # Create an OGR multistring object
        return ogr.CreateGeometryFromWkt(self.wkt_geometry)

    @property
    def wkt_geometry(self):
        return self.shapely_geometry.wkt

    def resample(self):
        # TODO: Need to implement this method
        pass

    def speed_percentiles(self, percentiles=None):
        speed_kmhr = self._dataframe['speed_kmhr']
        if percentiles is None:
            percentiles = [0.5]
        if not isinstance(percentiles, list):
            percentiles = [percentiles]
        return speed_kmhr.quantile(q=percentiles, interpolation='linear')


class MultiTrajectory(Trajectory):

    def __init__(self, multi_relocs):
        _multirelocs_df = multi_relocs.df()
        self.crs = _multirelocs_df.crs

        if _multirelocs_df.crs == CRS.WGS84:
            self._gdf = _multirelocs_df.groupby('subject_id').apply(self._create_multitraj).droplevel(['subject_id'])
        else:
            _relocs_df = _multirelocs_df.to_crs(CRS.WGS84)
            self._gdf = _multirelocs_df.groupby('subject_id').apply(self._create_multitraj).droplevel(['subject_id'])
            self._gdf = self._gdf.to_crs(self.crs)

        self.relocs_df = _multirelocs_df

    def _create_multitraj(self, df):
        df['_fixtime'] = df['fixtime'].shift(-1)
        df['_geometry'] = df['geometry'].shift(-1)
        return self._create_trajsegments(df[:-1])


class AnalysisParams:
    """ An abstract class that acts as a container for calculation parameters"""
    pass


class AnalysisResult:
    """An abstract class to act as a container for calculation results"""

    def __init__(self):
        self._analysis_start = dt.datetime.utcnow()
        self._analysis_end = dt.datetime.utcnow()
        self._errors = []

    @property
    def analysis_start(self):
        return self._analysis_start

    @analysis_start.setter
    def analysis_start(self, value):
        if type(value) is dt.datetime:
            self._analysis_start = value

    @property
    def analysis_end(self):
        return self._analysis_end

    @analysis_end.setter
    def analysis_end(self, value):
        if type(value) is dt.datetime:
            self._analysis_end = value

    @property
    def errors(self):
        return self._errors

    def add_error(self, value):
        self._errors.append(value)


class SpatialFeature:
    """
    A spatial geometry with an associated name and unique ID. Becomes a useful construct in several movdata calculations
    """

    def __init__(self, ogr_geometry=None, name='', unique_id=None):
        self._ogr_geometry = ogr_geometry
        self._name = name
        self._unique_id = unique_id or guid.uuid4()

    @property
    def name(self):
        return str(self._name)

    @property
    def unique_id(self):
        return self._unique_id

    @property
    def ogr_geometry(self):
        return self._ogr_geometry

    @property
    def wkt_geometry(self):
        return self._ogr_geometry.ExportToWkt()

    @property
    def shapely_geometry(self):
        return wkt.loads(self.wkt_geometry)


class Region:
    """
    A polygon with an associated name. Becomes a useful construct in several movdata calculations
    """

    def __init__(self, shapely_geometry=None, region_name='', unique_id='', crs=CRS.WGS84):
        self._shapely_geometry = shapely_geometry
        self._region_name = region_name
        self._unique_id = unique_id
        self._crs = crs

    @property
    def region_name(self):
        return str(self._region_name)

    @property
    def unique_id(self):
        return self._unique_id or guid.uuid4()

    @property
    def ogr_geometry(self):
        return ogr.CreateGeometryFromWkt(self.wkt_geometry)

    @property
    def wkt_geometry(self):
        return self.shapely_geometry.wkt

    @property
    def shapely_geometry(self):
        return self.shapely_geometry

    @property
    def crs(self):
        return self._crs

    @staticmethod
    def to_crs(self, crs):
        tranformer = pyproj_transformer(self.crs, crs)
        return ops.transform(tranformer.transform, self.shapely_geometry)


class PymetException(Exception):
    """ A custom exception type for movdata specific errors"""
    pass



import movdata
import numpy as np

class Cluster:
    """A cluster represents a group of points in both time and space.
    Adds functionality to the base relocs object """

    def __init__(self, relocs=None):
        self._relocs = relocs or movdata.base.Relocations()

    @property
    def centroid(self):
        return self._relocs.centroid

    @property
    def relocs(self):
        return self._relocs

    def add_fix(self, fix):
        self._relocs.add_fix(fix)

    @property
    def cluster_radius(self):
        """ The cluster radius is the largest distance between a point in the relocs and the
        centroid of the relocs"""

        cluster_radius = 0.0
        centroid = self.relocs.centroid
        for pnt in self.relocs.get_fixes():
            #calculate the distance between the centroid and the fix
            this_radius = pnt.geopoint.dist_to_point(centroid)
            if this_radius > cluster_radius:
                cluster_radius = this_radius

        return cluster_radius

    @property
    def cluster_std_dev(self):
        """The cluster standard deviation is the standard deviation of the radii from the centroid
        to each point making up the cluster"""
        _centroid = self.relocs.centroid
        _radii = []
        for _pnt in self.relocs.get_fixes():
            # calculate the distance between the centroid and the fix
            _radii.append(_pnt.geopoint.dist_to_point(_centroid))
        return np.std(_radii)

    def threshold_point_count(self, threshold_dist):
        """Counts the number of points in the cluster that are within a threshold distance of the geographic centre"""
        _clust_count = 0
        _centroid = self.relocs.centroid
        for _pnt in self.relocs.get_fixes():
            # calculate the distance between the centroid and the fix
            _this_radius = _pnt.geopoint.dist_to_point(_centroid)
            if _this_radius <= threshold_dist:
                _clust_count = _clust_count + 1
        return _clust_count

    def apply_threshold_filter(self, threshold_dist_meters=float("Inf")):
        _centroid = self.relocs.centroid
        if _centroid is not None:
            for _f in self.relocs._fixes: #Apply to the underlying array and not to get_fixes() array
                _this_radius = _f.geopoint.dist_to_point(_centroid)
                if _this_radius > threshold_dist_meters:
                    _f.junk_status=True


import datetime as dt
import pytz

import time
import json
import shapely
import geojson
import ee
import geopandas as gpd
import pandas as pd
import dask.dataframe as dd
# from signal import signal, SIGPIPE, SIG_DFL https://docs.python.org/3/library/signal.html#note-on-sigpipe

import logging
logger = logging.getLogger(__name__)

import os

ee_is_initialized = False

ee_initialization_expires = dt.datetime(1970, 1, 1, tzinfo=pytz.utc)
ee_initialization_ttl = dt.timedelta(seconds=86400)


def initialize_earthengine(key_dict):
    '''
    This takes a JSON key as a dict.
    :param key_dict:
    :return: a credentials object that can be used to initialize the earth engine library.
    '''
    global ee_is_initialized
    global ee_initialization_expires

    try:
        if ee_is_initialized and (dt.datetime.now(tz=pytz.utc) > ee_initialization_expires):
            logger.debug('Earth Engine has already been initialized.')
            return

        sac = ee.ServiceAccountCredentials(key_dict['client_email'], key_data=json.dumps(key_dict))
        ee.Initialize(credentials=sac)

    except Exception as e:
        logger.exception('Not able to initialize EarthEngine.')
        raise
    else:
        ee_is_initialized = True
        ee_initialization_expires = dt.datetime.now(tz=pytz.utc) + ee_initialization_ttl
        logger.debug('Earth Engine has been initialized.')


# useful stylizations
colourPalettes = {'precipitation': {'palette': ['00FFFF', '0000FF']}, }


def convert_millisecs_datetime(unix_time):
    # Define the UNIX Epoch start date
    epoch = dt.datetime(1970, 1, 1, 0, 0, 0)
    t = epoch + dt.timedelta(milliseconds=unix_time)
    return t


def convert_datetime_millisecs(input_time):
    epoch = dt.datetime(1970, 1, 1, 0, 0, 0) # Define the UNIX Epoch start date
    return int((input_time - epoch).total_seconds() * 1000)


def convertUnixTimeToDateTime(unixTime):
    t = dt.datetime.utcfromtimestamp(int(unixTime/1000))
    return t


def add_img_time(img):

    """ A function to add the date range of the image as one of its
    properties and the start and end values as new bands"""

    # unable to use the band rename() function here in Python whereas works in JS playground
    img = img.addBands(img.metadata('system:time_start'))
    img = img.addBands(img.metadata('system:time_end'))
    dr = ee.DateRange(ee.Date(img.get('system:time_start')),
                      ee.Date(img.get('system:time_end')))

    return img.set({'date_range': dr})


def fixes_to_featurecollection(fixes):

    """ A function to turn a relocations object into a GEE feature collection"""

    return ee.FeatureCollection([ee.Feature(ee.Geometry.Point(p.geopoint.X, p.geopoint.Y),
                                            {'fixtime': p.fixtime.replace(tzinfo=None), 'uuid': p.uuid})
                                 for p in fixes])


def gdf_to_featurecollection(gdf):

    feats = []
    for i, row in gdf.iterrows():
        ee_feat = ee.Feature(geojson_to_ee_geometry(shapely.geometry.mapping(row.geometry)), {'pdindex': row.name})
        feats.append(ee_feat)
    return ee.FeatureCollection(feats)


def geojson_to_ee_geometry(val):
    """
    A function to convert geojson to an ee.Geometry
    :param val: can either be a geojson formatted string or dict
    :return: ee.Geometry or None
    """
    result = None
    if isinstance(val, str):
        try:
            result = ee.Geometry(geojson.loads(val))
        except:
            logger.exception('Failed to read geojson from %s', val)
    elif isinstance(val, dict):
        result = ee.Geometry(val)

    return result


def geofile_to_gdf(geofile=''):
    """ Read a geofile into a geopandas dataframe. Supports formats used by fiona"""
    """ Note that there is an issue with Fiona such that if the encoding is not UTF-8 (e.g.,ANSI 1252)
    then the import will fail. The encoding is usually listed in a cpg file. If this happens simply delete
     the cpg file. The default for ArcGIS generated shapefiles is UTF-8"""
    return gpd.GeoDataFrame.from_file(geofile)


# Define a function that will extract image values from GEE.
def example_extract_ndvi_and_slope_data_function(feat):

    # NDVI
    ndvi_img = ee.ImageCollection('MODIS/MCD43A4_NDVI').select('NDVI').mean()  # get the mean NDVI value
    ndvi_reduce = ndvi_img.toArray().reduceRegion(ee.Reducer.toList(), feat.geometry(), 500.0)

    # Slope
    slope_reduce = ee.Algorithms.Terrain(ee.Image("USGS/SRTMGL1_003")). \
        select(['slope']).reduceRegion(ee.Reducer.toList(), feat.geometry(), 30.0)

    return {
        'Slope': slope_reduce,
        'NDVI': ndvi_reduce,
    }


def example_regional_albedo_anomaly_extract_function(feat):
    """
    :param feat: an earth engine feature
    :return: earth engine compute object
    """

    historical_start = '2010-01-01'
    start = '2015-01-01'
    end = '2020-11-01'
    scale = 5000

    img_coll = ee.ImageCollection('MODIS/006/MCD43C3').select('Albedo_BSA_vis')
    reference = img_coll.filterDate(historical_start, start).sort('system:time_start', False)
    ref_mean = reference.mean()

    '''
    Compute anomalies by subtracting the 2001-2010 mean from each image in a collection of 2011-2020 images.
    Copy the date metadata over to the computed anomaly images in the new collection.
    '''
    def tmp1(image):
        return image.subtract(ref_mean).set('system:time_start', image.get('system:time_start'))
        #reduceRegion(ee.Reducer.mean(), feat.geometry(), scale). \

    series = img_coll.filterDate(start, end).map(tmp1)

    series_reduce = series.toArray().reduceRegion(ee.Reducer.toList(), feat.geometry(), scale)

    # anomalySumImg = series.sum()

    # time0 = reference.first().get('system:time_start')
    # first = ee.List([ee.Image(0).set('system:time_start', time0).select([0], ['Albedo'])])
    #
    # '''This is a function to pass to Iterate(). As anomaly images are computed, add them to the list.'''
    #
    # def accumulate(image, inlist):
    #
    #     # Get the  latest cumulative anomaly image from the end of the list
    #     # with get(-1).Since the type of the list argument to the function is unknown, it needs to be cast to a List.
    #     # Since the return type of get() is unknown, cast it to Image.
    #     previous = ee.Image(ee.List(inlist).get(-1))
    #
    #     # Add the current anomaly to make a new cumulative anomaly image.
    #     added = image. \
    #         add(previous).set('system:time_start', image.get('system:time_start'))
    #
    #     # Return the list with the cumulative anomaly inserted.
    #     return ee.List(inlist).add(added)
    #
    # cumulative = ee.ImageCollection(ee.List(series.iterate(accumulate, first)))
    #
    # cumulative_reduce = cumulative.toArray().reduceRegion(ee.Reducer.mean(), feat.geometry(), scale)

    return feat.set({'img_vals': series_reduce})


def example_tracking_data_extract_data_function(feat):

    '''
     before_win: the amount of time in seconds to select before the fix's fixtime. Set to 0 for the fixtime only
     after_win: the amount of time in seconds to select after the fix's fixtime set to 0 for the fixtime
     limit: the number of images to select that are temporally close to the fixtime. Use zero for all images
     scale: analysis scale in meters
     img_coll: the image collection to use for the extraction
    '''

    before_win = 16 * 24 * 60 * 60
    after_win = 0
    limit = 1
    scale = 500.
    img_coll = ee.ImageCollection('MODIS/MCD43A4_NDVI').select('NDVI');


    fixtime = ee.Date(feat.get('fixtime'))
    t1 = fixtime.advance(-1 * before_win, 'second')
    t2 = fixtime.advance(after_win, 'second')

    # filter the image collection by geometry
    tmpColl = img_coll.filterDate(t1, t2)  # .filterBounds(feat.geometry())

    # calculate how far away in time the image time_start value is from the current fix's fixtime
    def temporal_proximity(img):
        # positive means img is before, negative means img is after
        temporal_dist = ee.Date(feat.get('fixtime')).difference(img.date(), 'second').abs()
        img = img.set({'temporal_dist': temporal_dist,})
        return img

    # Attempt 1 to filter based on date range
    # Only include images whose date range intersects the date range of the feature
    # tmpColl = ee.ImageCollection(ee.Algorithms.If(temporal_intersect,
    #                                               tmpColl.filter(ee.Filter.eq('temporal_intersects', True)),
    #                                               tmpColl))

    # Attempt 2 to filter based on date range
    # tmpColl = ee.ImageCollection(ee.Algorithms.If(temporal_intersect,
    #                                               tmpColl.filter(
    #                                               ee.Filter.dateRangeContains('date_range', fixtime)), tmpColl))

    tmpColl = tmpColl.map(temporal_proximity).sort('temporal_dist', True).limit(limit)

    # Reduce
    tmp_reduce = tmpColl.toArray().reduceRegion(ee.Reducer.toList(), feat.geometry(), scale)

    return feat.set({'img_vals': tmp_reduce})


def parallelize_dataframe(df, func, n_cores=4):
    """ Input needs to be a plain pandas dataframe with a lat and lon column since Dask  
    can't interpret the Geometry column """

    # Using dask
    def dask_func(df1):
        gdf = gpd.GeoDataFrame(df1, geometry=gpd.points_from_xy(df1.lon, df1.lat), crs="EPSG:4326")
        return extract_img_values(gdf=gdf, img_calc=func, chunk_size=2000)

    ddf = dd.from_pandas(df, npartitions=n_cores)
    result = ddf.map_partitions(dask_func).compute()

    return result


def extract_img_values(gdf=None, img_calc=None, chunk_size=2000):
    """
    A function to run an image extraction function for each spatial geometry available in a geopandas dataframe

    :param gdf: a geopandas dataframe. The 'geometry' column can be any type pf geometry (point/line/polygon)
    :param img_calc: a function that will be used to extract image values intersecting the spatial geometry
    """

    def map_features(feat):
        return feat.set({img_calc.__name__: img_calc(feat)})

    results = dict()

    for i in range(0, len(gdf.index), chunk_size):
        try:
            success = False
            logger.info('Analyzing features: ' + str(i) + ' to ' + str(i + chunk_size - 1))
            chunk = gdf[i:i + chunk_size]

            # Build the relocations into a GEE feature collection
            feat_coll = gdf_to_featurecollection(chunk)
            while not success:
                try:
                    # Run the map function over all features
                    result = feat_coll.map(map_features, True).getInfo()

                    # Get the results for each image calculation
                    for f in result.get('features'):
                        results[f.get('properties').get('pdindex')] = \
                            f.get('properties').get(img_calc.__name__, None)
                    success = True

                except Exception as ex:
                    logger.error(str(ex))
                    logger.info('Re-trying in 5 seconds...')
                    time.sleep(5)  # Pause for 5 seconds before trying again
        except Exception as ex:
            logger.error(str(ex))

    if results:
        # Convert the results into a dataframe
        results = pd.DataFrame.from_dict(data=results, orient='index')
    else:
        results = None

    return results


# ------------------------- todo: eventually deprecate everything below ----------------------------------------------------------


def extract_temporal_img_values(fixes=None, img_coll=None, calc_name='gee_calc',
                                before_win=16*24*60*60, after_win=0, limit=1,  # temporal_intersect=True,
                                scale=500.0, chunk_size=20):

    """
    A function to filter the image collection to just those images that either intersect
    the fixtime daterange, or are the closest temporally, and intersect the geometry of the
     feature and then extract the image pixel values.

    fixes: a list of fix objects
    img_coll: the GEE image collection
    calc_name: the name to give this particular calculation. It will be the key to use to look up results in the
    fix's gee_result dictionary
    before_win: the amount of time in seconds to select before the fix's fixtime. Set to 0 for the fixtime only
    after_win: the amount of time in seconds to select after the fix's fixtime set to 0 for the fixtime
    nearest: the number of images to select that are temporally close to the fixtime. Use zero for all images
    temporal_intersect: require that the image's daterange intersects with the fixe's daterange
    scale: the pixel size in meters at which to resample the imagery
    chunk_size: the number of features that should be processed at once within GEE. This helps limiting memory errors
    """

    # Exit if the fixes list are null or empty
    if (fixes is None) or (fixes == []):
        return

    # Add date ranges to all the images in the collection
    img_coll = img_coll.map(add_img_time)

    # Define the function to map over each feature
    def map_features(feat):

        fixtime = ee.Date(feat.get('fixtime'))
        t1 = fixtime.advance(-1 * before_win, 'second')
        t2 = fixtime.advance(after_win, 'second')
        feat_dr = ee.DateRange(t1, t2)

        # filter the image collection by geometry
        # tmpColl = img_coll.filterBounds(feat.geometry())
        tmpColl = img_coll.filterDate(t1, t2)  # .filterBounds(feat.geometry())


        # calculate how far away in time the image time_start value is from the current fix's fixtime
        # and whether the image timespan intersects the fix's designated time window
        def temporal_proximity(img):
            # positive means img is before, negative means img is after
            temporal_dist = ee.Date(feat.get('fixtime')).difference(img.date(), 'second').abs()
            # temporal_intersects = feat_dr.intersects(img.get('date_range'))
            # intersects_val = ee.Algorithms.If(temporal_intersects, ee.Number(1), ee.Number(0))
            img = img.set({'temporal_dist': temporal_dist,
                           # 'temporal_intersects': temporal_intersects
                           })
            # img = img.addBands([img.metadata('temporal_dist'), img.metadata('intersects_val')])
            return img

        # Attempt 1 to filter based on date range
        # Only include images whose date range intersects the date range of the feature
        # tmpColl = ee.ImageCollection(ee.Algorithms.If(temporal_intersect,
        #                                               tmpColl.filter(ee.Filter.eq('temporal_intersects', True)),
        #                                               tmpColl))

        # Attempt 2 to filter based on date range
        # tmpColl = ee.ImageCollection(ee.Algorithms.If(temporal_intersect,
        #                                               tmpColl.filter(
        #                                               ee.Filter.dateRangeContains('date_range', fixtime)), tmpColl))

        tmpColl = tmpColl.map(temporal_proximity).sort('temporal_dist', True).limit(limit)

        # Reduce
        tmp_reduce = tmpColl.toArray().reduceRegion(ee.Reducer.toList(), feat.geometry(), scale)

        return feat.set({'img_vals': tmp_reduce})

    for i in range(0, len(fixes), chunk_size):
        try:
            print('Analyzing features: ', str(i), ' to ', str(i + chunk_size-1))
            chunk = fixes[i:i + chunk_size]

            # Build the relocations object into a GEE feature collection
            feat_coll = fixes_to_featurecollection(chunk)

            # success = False
            # while success is False:
            #     # The while loop lets us try over if the GEE servers didn't respond and we get a server timeout error.
            #     # Because of lazy execution the servers may not be finished running a calc but the next time you
            #     # try the result is ready.
            #     # ToDo: figure out how to discern between a result not ready error and some other error type


            # Run the map function over all features
            result = feat_coll.map(map_features).getInfo()
            # success = True

            # print(result.get('features'))

            # Create a dictionary using the fix ID values as the key:
            tmp_results = dict()

            for feat in result.get('features'):
                img_array = feat.get('properties').get('img_vals').get('array', None)
                if img_array:
                    tmp_results[feat.get('properties').get('uuid')] = img_array[0]

            # Set the results for each input fix
            for fix in chunk:
                fix.additional_data[calc_name] = tmp_results.get(fix.uuid)
                # print(fix.uuid, fix.properties[calc_name])

        except Exception as ex:
            print(ex)
            # print('Re-trying in 5 seconds...')
            # time.sleep(5)  # Pause for 5 seconds before trying again


# In use by DAS_analyzers
# ToDO: have changed this from relocs to fixes in the params - need to update DAS
# ToDo: have updated from calculating the mean to returning the raw output from EE - need to update DAS
def extract_point_values_from_image(fixes, img_name='', band_name='b1', scale=500.0):
    """ A function to extract values from an image at relocations """

    # Exit if the relocs object is null
    if fixes is None:
        return

    # Build the relocations into a GEE feature collection
    feat_coll = fixes_to_featurecollection(fixes)

    try:
        img = ee.Image(img_name)

        # dictionary result of the reduction
        result = img.reduceRegion(ee.Reducer.toList(), feat_coll.geometry(), scale).getInfo()
        # ToDO: how to link back to original feature ID?
        return result

        # if result is not None:
        #     raw_vals = result.get(band_name)
        #     if raw_vals is not None:
        #         return np.mean(raw_vals)
    except Exception as ex:
        print(ex)
        return


def extract_point_values_from_image2(fixes=None, ee_img=None, calc_name='gee_calc', scale=500.0, chunk_size=500):
    """ A function to extract values from an image at relocations """

    # Exit if the relocs object is null
    if fixes is None:
        return

    # Build the relocations into a GEE feature collection
    feat_coll = fixes_to_featurecollection(fixes)

    def map_features(feat):

        # Reduce
        tmp_reduce = ee_img.reduceRegion(ee.Reducer.toList(), feat.geometry(), scale)
        return feat.set({'img_val': tmp_reduce})


    for i in range(0, len(fixes), chunk_size):
        try:
            print('Analyzing features: ', str(i), ' to ', str(i + chunk_size-1))
            chunk = fixes[i:i + chunk_size]

            # Build the relocations object into a GEE feature collection
            feat_coll = fixes_to_featurecollection(chunk)

            # Run the map function over all features
            result = feat_coll.map(map_features).getInfo()
            # success = True

            # print(result.get('features'))

            # Create a dictionary using the fix ID values as the key:
            tmp_results = dict()

            for feat in result.get('features'):
                img_val = feat.get('properties').get('img_val')
                if img_val:
                    tmp_results[feat.get('properties').get('uuid')] = img_val

            # Set the results for each input fix
            for fix in chunk:
                fix.additional_data[calc_name] = tmp_results.get(fix.uuid)

        except Exception as ex:
            print(ex)
            return


def ReduceRegionByDatetimes(image_stack,
                            region,
                            region_scale,
                            ee_region_reducer_type,
                            ee_stack_reducer_type,
                            band_name,
                            datetimes=[]):

    """A function to perform a series of image stack reductions on the chosen band
     within the given region and for the given datetime tuples"""

    # Get the GPM dataset
    imgColl = ee.ImageCollection(image_stack).select(band_name).sort('system:time_start', False)

    # Convert the datetime tuples into a feature collection
    featColl = ee.FeatureCollection([ee.Feature(None, {'startDate': x[0], 'endDate': x[1]}) for x in datetimes])

    def mapFunc(feat):
        reduc = imgColl.filterDate(feat.get('startDate'), feat.get('endDate')).reduce(
            ee_stack_reducer_type).reduceRegion(ee_region_reducer_type, region.geometry(), region_scale)
        return feat.set({'reduceVal': reduc})

    return [(convert_millisecs_datetime(x.get('properties').get('startDate').get('value')),
             x.get('properties').get('reduceVal').get(band_name + '_' + ee_stack_reducer_type))
            for x in featColl.map(mapFunc).getInfo()['features']]



import math
import os
from osgeo import ogr
from osgeo import osr
import random
import rasterio
from rasterio.mask import mask
import geopandas as gpd
import pyproj
from shapely.ops import transform
from shapely.geometry import LineString, Point



def km_to_degrees(km):
    """Quick and dirty conversion from kilometers to degrees"""
    rads = km / 6371.0
    return rads*180.0/math.pi
    #return math.degrees(rads)



def degrees_to_km(degrees):
    rads = math.radians(degrees)
    return rads * 6371.0


def rect_coord_geometry(center=None, side_len_km=1.0, srs=None):

    if srs is None:
        # create the spatial reference, WGS84
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(4326)

    def rect_from_center(xc, yc, length_km):
        """ Generate rectangle corner coordinates based on the side length and center point """
        s = km_to_degrees(length_km / 2)
        p1 = (xc - s, yc + s)
        p2 = (xc - s, yc - s)
        p3 = (xc + s, yc - s)
        p4 = (xc + s, yc + s)
        return p1, p2, p3, p4

    # Create ring
    ring = ogr.Geometry(ogr.wkbLinearRing)

    for pt in rect_from_center(xc=center[0], yc=center[1], length_km=side_len_km):
        ring.AddPoint(pt[0], pt[1])

    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AssignSpatialReference(srs)
    poly.AddGeometry(ring)

    return poly


def generate_AOI_shapefile(AOI_center_coords=None, side_len_km=1.0,
                           output_location='', output_name='AOI_sites.shp'):
    """Create an output shapefile with equal sized AOI rectangle polyons
     based on the input AOI_center_coords.
    AOI_center_coords are lon/lat pairs"""

    # set up the shapefile driver
    driver = ogr.GetDriverByName("ESRI Shapefile")

    # create the data source
    data_source = driver.CreateDataSource(os.path.join(output_location, output_name))

    # create the spatial reference, WGS84
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(4326)

    # create the layer
    layer = data_source.CreateLayer("sites", srs, ogr.wkbPolygon)

    # Create fields
    field_name = ogr.FieldDefn("Name", ogr.OFTString)
    field_name.SetWidth(24)
    layer.CreateField(field_name)

    for i, pnt in enumerate(AOI_center_coords):
        try:
            # create the feature
            feature = ogr.Feature(layer.GetLayerDefn())

            # Set the attributes using the values from the delimited text file
            feature.SetField("Name", 'Sample Site ' + str(i))

            aoi_geo = rect_coord_geometry(center=pnt, side_len_km=side_len_km)

            # Create the point from the Well Known Txt
            shp = ogr.CreateGeometryFromWkt(aoi_geo.ExportToWkt())

            # Set the feature geometry using the point
            feature.SetGeometry(shp)

            # Create the feature in the layer (shapefile)
            layer.CreateFeature(feature)

        except:
            print('An error occurred')
            pass


def assess_containment(geopoint=None, regions=None):
    """Pass back an array of all the regions IDs that contain the given geopoint"""
    containment = []

    if geopoint is not None:
        if regions is None:
            return []

        for region in regions:
            if type(region) is SpatialFeature:
                contains = region.ogr_geometry.Contains(geopoint.ogr_geometry)
                if contains is True:
                    containment.append(region.unique_id)

    return containment


def generate_csr_points_region(region=None, num=1000):

    # check if the region is a Region object
    if type(region) is not Region:
        raise Exception('Input polygon is not of type movdata.base.Region')

    poly = region.ogr_geometry

    env = poly.GetEnvelope()
    x_min, x_max, y_min, y_max = env[0], env[1], env[2], env[3]

    list_of_points = []
    counter = 0
    while counter < num:
        pnt = ogr.Geometry(type=ogr.wkbPoint)
        pnt.AddPoint(random.uniform(x_min, x_max), random.uniform(y_min, y_max), 0.)
        pnt.AssignSpatialReference(SRS.WGS84())

        if poly.Contains(pnt):
            list_of_points.append(pnt)
            counter += 1
    return list_of_points


def habiba_trajectory():
    """Create a trajectory based on two weeks of movement of the elephant Habiba in Samburu National Reserve"""

    datafile = r'../examples/sample_data/habiba.csv'
    movdata = pd.read_csv(datafile, sep=',', header=0)
    movdata['Fixtime'] = pd.to_datetime(movdata['Fixtime'], format="%Y-%m-%d %H:%M:%S", utc=True)
    # Create a trajectory from the dataframe
    relocs = Relocations([Fix(GeoPoint(i.Longitude, i.Latitude, 0), i.Fixtime) for i in movdata.itertuples()], 'Habiba')
    traj = Trajectory(relocs)
    return traj


def samburu_point():
    """Create a random OGR point in Samburu National Reserve, Kenya"""
    point = ogr.Geometry(type=ogr.wkbPoint)
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(4326)
    point.AddPoint(x=37.52, y=0.58, z=0)
    point.AssignSpatialReference(srs)
    return point


def ewaso_river():
    """Create the OGR geometry for the Ewaso Nyiro river flowing through Samburu National Reserve, Kenya"""
    shapefile = r'../examples/Data/spatial_data/EwasoNyiroRiver.shp'
    driver = ogr.GetDriverByName("ESRI Shapefile")
    data_source = driver.Open(shapefile, 0)
    layer = data_source.GetLayer()
    geoms = []
    for feature in layer:
        geom = feature.GetGeometryRef()
        geoms.append(geom)
    return geoms[0]


def samburu_polygon():
    """Create a random polygon based on a random point in Samburu National Reserve, Kenya"""
    pnt = samburu_point()
    return rect_coord_geometry(center=(pnt.GetX(), pnt.GetY()), side_len_km=km_to_degrees(10.0))


def reduce_region(gdf=None, rasters=None, reduce_func=None):
    """
    A function to apply the reduce_func to the values of the pixels within each of the rasters for every
    shape within the input geopandas dataframe 'geometry' column
    :param gdf: geopandas dataframe
    :param rasters: a list of raster names
    :return: None
    """
    # define valid region reduce functions
    #reduce_funcs = [np.mean, np.sum, np.min, np.max, np.percentile, np.count_nonzero]

    try:
        # check for correct input types
        assert type(gdf) is gpd.geodataframe.GeoDataFrame
        assert type(rasters) is list
        #assert reduce_func in reduce_funcs  # ToDo: better way to test the type of reduce function?

        def tmp(r):
            result = dict()
            for raster_name in rasters:
                print('Processing: ', raster_name, ' and ', r.name)
                try:
                    with rasterio.open(raster_name) as src:
                        # src is an ndarray with dimensions (num_band, num_rows, num_cols)
                        # out_img is a masked ndarray that uses the same NoData value as
                        # the input raster (defaults to 0)
                        project = pyproj.Transformer.from_crs(gdf.crs, src.crs, always_xy=True).transform
                        proj_shp = transform(project, r['geometry'])
                        out_img, out_transform = mask(src, [proj_shp], filled=False)

                        # apply the input reduce_func to the masked ndarray.
                        # Note that this also performs a stack reduce
                        # if there is more than one band
                        result[raster_name] = reduce_func(out_img.compressed())

                except Exception as ex:
                    print(ex)
            return result

        gdf['reduce_region'] = gdf.apply(tmp, axis=1)

    except Exception as ex:
        print(ex)


import movdata.eetools as eetools
import ee
import sklearn.mixture as gauss
from sklearn.preprocessing import LabelEncoder
import pandas as pd
import datetime as dt
import logging
from scipy.stats import norm
import numpy as np

logger = logging.getLogger(__name__)


class Seasons:

    def __init__(self, start=None, end=None,
                 season_n=2, season_labels=None,
                 img_col='MODIS/MCD43A4_006_NDVI', band='NDVI',
                 analysis_scale_m=500, analysis_geo=None, analysis_date_chunks=3,
                 calibration_func=None):
        self._start = start
        self._end = end
        self._season_n = season_n
        self._season_labels = season_labels
        self._img_col = img_col
        self._band = band
        self._analysis_scale_m = analysis_scale_m
        self._analysis_geo = analysis_geo
        self._seasonal_cuts = None
        self._vals = None
        self._analysis_date_chunks = analysis_date_chunks  # ~ 2500 image collection limit in GEE
        self._calibration_func = calibration_func

    # ToDo: rewrite this function using a geopands DF as an input to handle multi-regions. And also write it using the
    # eetools.extract_img_values paradigm where the function 'extract_seasonal_values' is provided as input to the
    # eetools.extract_img_values function

    def extract_seasonal_vals(self, ee_geo=None, scale=500):

        """
        :param ee_geo: ee_geo is an EE Geometry type
        :param scale:
        :return:
        """

        # To-Do: define an identity calibration function if none exists
        # img_coll = ee.ImageCollection(self._img_coll).select(self._band).map(self.unityfunc).toArray()

        secs = self._end.timestamp() - self._start.timestamp()
        chunk_size = secs // self._analysis_date_chunks
        date_slices = []
        cur_timestamp = self._start.timestamp()
        while cur_timestamp < self._end.timestamp():
            date_slices.append(cur_timestamp)
            cur_timestamp += chunk_size
        date_slices.append(self._end.timestamp() + 1)

        result_df = []

        for i in range(len(date_slices) - 1):

            start = dt.datetime.fromtimestamp(date_slices[i])
            end = dt.datetime.fromtimestamp(date_slices[i + 1] - 1)
            print('Processing date range: ', start, end)

            # get the image stack
            img_coll = ee.ImageCollection(self._img_col).select(self._band). \
                filterBounds(self._analysis_geo).filter(ee.Filter.date(start, end)). \
                sort('system:time_start', False)

            # add time info
            tmpColl = img_coll.map(eetools.add_img_time).toArray()

            # extract the pixel values falling within the region and for the given dates
            ee_result = tmpColl.reduceRegion(ee.Reducer.toList(), self._analysis_geo, self._analysis_scale_m).getInfo()

            vals = pd.DataFrame([[x[0], eetools.convertUnixTimeToDateTime(x[1])] for y in ee_result.get('array')
                                 for x in y], columns=['val', 'datetime'])

            if vals is not None:
                if self._calibration_func is not None:
                    vals['val'].apply(self._calibration_func)
                result_df.append(vals)

        self._vals = pd.concat(result_df, axis=0).sort_values(by='datetime')

        return self._vals

    def determine_seasonal_cut_values(self):

        try:
            if self._vals is not None:
                distr = gauss.GaussianMixture(n_components=self._season_n, max_iter=500)
                vals = self._vals['val'].values.reshape(-1, 1)
                distr.fit(vals)
                mu_vars = np.array(sorted(zip(distr.means_.flatten(),
                                              distr.covariances_.flatten()), key=lambda x: x[0]))
                print('mu_vars', mu_vars)
                cuts = []
                x = np.linspace(0, 1.0, 1000)
                for i in range(1, len(mu_vars)):
                    cuts.append(np.max(x[tuple([norm.sf(x, loc=mu_vars[tuple([i - 1, 0])],
                                                  scale=np.sqrt(mu_vars[tuple([i - 1, 1])])) >
                                          norm.cdf(x, loc=mu_vars[tuple([i, 0])],
                                                   scale=np.sqrt(mu_vars[tuple([i, 1])]))])]))
                cuts.append(float("inf"))
                cuts.append(float("-inf"))
                self._seasonal_cuts = sorted(cuts)

            else:
                logger.exception('Could not run EM algorithm on null vals array.')
        finally:
            return self._seasonal_cuts

    def determine_seasonal_windows(self, reducer_func=np.mean):

        result = None

        try:
            self.extract_seasonal_vals()
            self.determine_seasonal_cut_values()

            if self._seasonal_cuts is None:
                logger.exception('Could not determine_seasonal_windows on null cut values array.')

            if self._vals is None:
                logger.exception('Could not determine_seasonal_windows on null values array.')

            def g(group):
                group['reduce_val'] = group['val'].agg(reducer_func)
                return group.iloc[0]  # return first row only

            df = self._vals.groupby('datetime').apply(g)
            df.drop(labels=['val'], axis=1, inplace=True)

            season_labels = self._season_labels
            if season_labels is None:
                season_labels = ['season_' + str(x) for x in range(len(self._seasonal_cuts))]

            df['season'] = pd.cut(df['reduce_val'], bins=self._seasonal_cuts, labels=season_labels)

            enc = LabelEncoder()
            df['season_code'] = enc.fit_transform(df['season'])

            '''Adapted from https://stackoverflow.com/questions/26911851/
            how-to-use-pandas-to-find-consecutive-same-data-in-time-series'''
            df['Unique Season'] = (df['season_code'].diff(1) != 0).astype('int').cumsum()
            df['end'] = df['datetime'].shift(-1)

            # data = OrderedDict([('start', df.groupby('Unique Season')['datetime'].first()),
            #                     ('end', df.groupby('Unique Season')['end'].last()),
            #                     ('season', df.groupby('Unique Season')['season'].first())])
            #
            # result = pd.DataFrame(data, columns=data.keys())

            result = pd.concat([pd.Series([grp['datetime'][0] for name, grp in df.groupby('Unique Season')]),
                                pd.Series([grp['datetime'][-1] for name, grp in df.groupby('Unique Season')]),
                                pd.Series([name for name, _ in df.groupby('Unique Season')]),
                                pd.Series(grp['season'][0] for name, grp in df.groupby('Unique Season'))], axis=1)
            result.columns = ['start', 'end', 'unique_season', 'season']
            result.set_index('unique_season', inplace=True)

        except Exception as e:
            print(e)

        finally:
            return result



import movdata.base as base
import datetime as dt
import os
import numpy as np
import fiona
import typing
import rasterio
import rasterio.features
import pandas as pd
import geopandas as gp
from dataclasses import dataclass, field
from shapely.geometry import shape, mapping
from shapely.geometry.multipolygon import MultiPolygon


@dataclass
class PercentileAreaProfile(base.AnalysisParams):
    input_raster: typing.Union[str, bytes, os.PathLike]
    percentile_levels: typing.List = field(default_factory=[50.])
    subject_id: str = ''


class PercentileAreaAnalysisResults(base.AnalysisResult):

    def __init__(self,
                 percentile=100.0,
                 area=0.0,
                 geo=None):
        super().__init__()
        self._percentile = percentile
        self._area = area
        self._geo = geo

    @property
    def percentile(self):
        return self._percentile

    @property
    def area(self):
        return self._area

    @property
    def geo(self):
        return self._geo


class PercentileArea:

    @staticmethod
    def _multipolygon(shapes, percentile):
        return MultiPolygon([shape(geom) for geom, value in shapes if value == percentile])

    @classmethod
    def calculate_percentile_area(cls, profile: PercentileAreaProfile):

        # create result object
        result = PercentileAreaAnalysisResults()
        result.analysis_start = dt.datetime.utcnow()

        try:
            assert type(profile) is PercentileAreaProfile
            shapes = []

            # open raster
            with rasterio.open(profile.input_raster) as src:
                crs = src.crs.to_wkt()

                for percentile in profile.percentile_levels:
                    data_array = src.read(1).astype(np.float32)

                    # Mask no-data values
                    data_array[data_array == src.nodata] = np.nan

                    # calculate percentile value
                    percentile_val = np.percentile(data_array[~np.isnan(data_array)], 100.0 - percentile)

                    # TODO: make a more explicit comparison for less than and greater than

                    # Set any vals less than the cutoff to be nan
                    data_array[data_array < percentile_val] = np.nan

                    # Mask any vals that are less than the cutoff percentile
                    data_array[data_array >= percentile_val] = percentile

                    shapes.extend(rasterio.features.shapes(data_array, transform=src.transform))

            data = [[profile.subject_id, percentile, cls._multipolygon(shapes, percentile)]
                    for percentile in sorted(profile.percentile_levels, reverse=True)]
            df = pd.DataFrame(data, columns=['subject_id', 'percentile', 'geometry'])

            return gp.GeoDataFrame(df, geometry=df.geometry, crs=crs)

        except Exception as e:
            result.add_error(e)





from datetime import datetime as dt
from datetime import timedelta as td
from dateutil import relativedelta as rd


class MovingWindow:
    def IterateWindow(self, traj, window_dates, func):
        pass


class Timespan(object):
    """An base class for either TimespanUnits or TimeSpans. Represents a window of time
    where the starting datetime is less than, or equal to the ending
    datetime."""

    def __init__(self):
        """ Timepsan constructor is meant to be overridden in subclasses"""
        self._start = dt.utcnow()
        self._end = dt.utcnow()

    def length_timedelta(self):
        # Return the datetime.timedelta between the two datetimes
        return self._end - self._start

    # Return the length of the timespan in seconds
    def length_timespan_seconds(self):
        return self.length_timedelta().total_seconds()

    def slide(self, timespan, amount=1):
        """ Slide the current timespan forward by the timespan amount. If the timespan object is a TimespanCustom
        type then the numtimes is in seconds"""
        self._start = timespan.add_time(self._start, amount)
        self._end = timespan.add_time(self._end, amount)

    @classmethod
    def add_time(cls, input_date, amount):
        """ Needs to be overridden in the implementing subclass for this function to do anything useful
        """
        pass


class TimespanUnit(Timespan):
    """ A Timepsan but that has a fundamental unit of measure (e.g., seconds, minutes, hours, days, weeks, months,
     years, etc."""

    def __init__(self, start, length=1):
        super().__init__()
        self.length = length
        self.start = start
        self.end = self.add_time(start, length)

    @classmethod
    def add_time(cls, input_date, amount):
        super().add_time(input_date, amount)

    def round_forwards(self):
        # Round backwards, then add a full X unit to the rounded time
        self.start = self.round_time_backwards(self.start)
        self.start = self.add_time(self.start, 1)
        self.end = self.add_time(self.start, 1)
        for i in range(1, self.length-1):
            self.end = self.add_time(self.end, 1)

    def round_backwards(self):
        # Round backwards, then add a full X unit to the rounded time
        self.start = self.round_time_backwards(self.start)
        self.end = self.add_time(self.start, 1)
        for i in range(1, self.length-1):
            self.end = self.add_time(self.end, 1)

    def getLength(self):
        """ return the length of the timespanunit in units"""
        return self.length

    @classmethod
    def round_time_forwards(cls, input_dt):
        # Round backwards, then add a full unit to the rounded time
        d1 = cls.round_time_backwards(input_dt)
        return cls.add_time(d1, 1)

    @classmethod
    def round_time_backwards(cls, input_dt):
        """Function will adjust the time window so that the start aligns with the earlier, whole TimespanUnit
                 unit of measure."""
        pass


class Timespan_Second(TimespanUnit):
    """A timespan represented in units seconds"""

    @classmethod
    def add_time(cls, input_date, amount):
        super().add_time(input_date, amount)
        return input_date + rd.relativedelta(seconds=+amount)

    @classmethod
    def round_time_backwards(cls, input_dt):
        """
        :param input_dt: the time to round to earliest second
        :return: datetime
        """
        super().round_time_backwards(input_dt)
        return input_dt - td(microseconds=input_dt.microsecond)


class Timespan_Minute(TimespanUnit):
    """A timespan represented in minutes"""

    @classmethod
    def add_time(cls, input_date, amount):
        super().add_time(input_date, amount)
        return input_date + rd.relativedelta(minutes=+amount)

    @classmethod
    def round_time_backwards(cls, input_dt):
        """
        :param input_dt: the time to round to earliest minute
        :return: datetime
        """
        super().round_time_backwards(input_dt)
        # round seconds to the nearest minute
        return input_dt - td(seconds=input_dt.second,
                             microseconds=input_dt.microsecond)


class Timespan_Hour(TimespanUnit):
    """A timespan represented in hours"""

    @classmethod
    def add_time(cls, input_date, amount):
        super().add_time(input_date, amount)
        return input_date + rd.relativedelta(hours=+amount)

    @classmethod
    def round_time_backwards(cls, input_dt):
        """
        :param input_dt: the time to round backwards
        :return: datetime
        """
        super().round_time_backwards(input_dt)
        # round minutes to the nearest hour
        return input_dt - td(minutes=input_dt.minute,
                             seconds=input_dt.second,
                             microseconds=input_dt.microsecond)


class Timespan_Day(TimespanUnit):
    """A timespan represented in days"""

    @classmethod
    def add_time(cls, input_date, amount):
        super().add_time(input_date, amount)
        return input_date + rd.relativedelta(days=+amount)

    @classmethod
    def round_time_backwards(cls, input_dt):
        """
        :param input_dt: the time to round backwards
        :return: datetime
        """
        super().round_time_backwards(input_dt)
        # round hours to the nearest day
        return input_dt - td(days=input_dt.hour,
                             minutes=input_dt.minute,
                             seconds=input_dt.second,
                             microseconds=input_dt.microsecond)


class Timespan_Week(TimespanUnit):
    """A timespan represented in weeks"""

    @classmethod
    def add_time(cls, input_date, amount):
        super().add_time(input_date, amount)
        return input_date + rd.relativedelta(weeks=amount)

    @classmethod
    def round_time_backwards(cls, input_dt):
        """
        :param input_dt: the time to round backwards
        :return: datetime
        """
        super().round_time_backwards(input_dt)

        # round day of the week to the nearest monday
        return input_dt - td(days=input_dt.weekday(),
                             hours=input_dt.hour,
                             minutes=input_dt.minute,
                             seconds=input_dt.second,
                             microseconds=input_dt.microsecond)


class Timespan_Month(TimespanUnit):
    """A timespan represented in months"""

    @classmethod
    def add_time(cls, input_date, amount):
        super().add_time(input_date, amount)
        return input_date + rd.relativedelta(months=+amount)

    @classmethod
    def round_time_backwards(cls, input_dt):
        """
        :param input_dt: the time to round backwards
        :return: datetime
        """
        super().round_time_backwards(input_dt)

        # round to the start of the month
        return input_dt - td(days=input_dt.day-1,
                             hours=input_dt.hour,
                             minutes=input_dt.minute,
                             seconds=input_dt.second,
                             microseconds=input_dt.microsecond)


class Timespan_Year(TimespanUnit):
    """A timespan represented in years"""

    @classmethod
    def add_time(cls, input_date, amount):
        super().add_time(input_date, amount)
        return input_date + rd.relativedelta(years=+amount)

    @classmethod
    def round_time_backwards(cls, input_dt):
        """
        :param input_dt: the time to round backwards
        :return: datetime
        """
        super().round_time_backwards(input_dt)

        # round to the start of the year
        return dt(year=input_dt.year, month=1, day=1, hour=0, minute=0, second=0, microsecond=0)


class Timespan_Custom(Timespan):
    """A class representing an arbitrary span of time"""

    def __init__(self, start, end):
        super().__init__()
        self.start = start
        self.end = end
        self.length = self.length_timespan_seconds() #The length of a custom timespan is measured in seconds

    @classmethod
    def add_time(cls, input_date, amount):
        super().add_time(input_date, amount)
        return input_date + rd.relativedelta(seconds=+amount)


class DatetimeGenerator:

    def step(self, timespan, latest_dt):
        """Generate a list of datetime tuples """
        outList = []

        while timespan.start_dt <= latest_dt:
            outList.append((timespan.start_dt, timespan.end_dt))
            timespan.slide(timespan, timespan.length_km)

        return outList

    # Todo
    def slide(self, base_ts, slide_ts, amount):
        pass



from __future__ import absolute_import, division, print_function, unicode_literals
from .version import __version__
from movdata import base, cluster, datanalyzer, dataschedule, geofence, kde_range, movingwindow, utils
from movdata import proximity, raster

__all__ = [base, cluster, datanalyzer, dataschedule, geofence, kde_range, movingwindow,
           utils, proximity, raster]





import movdata.base
import movdata.utils
import math
import datetime as dt


class ProximityAnalysisParams(movdata.base.AnalysisParams):
    """
    A class to hold parameters needed for the proximity calculation.
    spatial_features: a list of ogr geometries to run against each trajectory
    """

    def __init__(self, spatial_features=None):
        self._spatial_features = spatial_features or []

    @property
    def spatial_features(self):
        return self._spatial_features


class ProximityAnalysisResult(movdata.base.AnalysisResult):
    """ A class to hold the results of a proximity analysis"""

    def __init__(self, proximity_events=None):
        self._proximity_events = proximity_events or []

    @property
    def proximity_events(self):
        return self._proximity_events

    def add_proximity_event(self, proximity_event=None):
        if proximity_event is not None:
            self._proximity_events.append(proximity_event)


class ProximityEvent:

    """ Class to store the result of a single proximity event"""

    def __init__(self, subject_id='',
                 subject_speed=0.0,
                 subject_travel_heading=0.0,
                 proximity_distance_meters=math.inf,
                 proximal_fix=None,
                 spatial_feature_id='',
                 spatial_feature_name=''):

        self._subject_id = subject_id
        self._proximal_fix = proximal_fix
        self._spatial_feature_id = spatial_feature_id
        self._spatial_feature_name = spatial_feature_name
        self._subject_speed_kmhr = subject_speed
        self._subject_heading = subject_travel_heading
        self._proximity_distance_meters = proximity_distance_meters

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def subject_speed_kmhr(self):
        return self._subject_speed_kmhr

    @property
    def subject_heading(self):
        return self._subject_heading

    @property
    def proximal_fix(self):
        return self._proximal_fix

    @property
    def proximity_distance_meters(self):
        return self._proximity_distance_meters

    @property
    def spatial_feature_id(self):
        return self._spatial_feature_id

    @property
    def spatial_feature_name(self):
        return self._spatial_feature_name


class ProximityAnalysis:

    @classmethod
    def calc_proximity_events(cls, proximity_analysis_params=None, trajectories=None):
        """
        :param proximity_analysis_params:
        :param trajectories:
        :return:
        """

        trajectories = trajectories or []

        # Test to make sure the calculation parameters are not None and the correct type
        assert type(proximity_analysis_params) is ProximityAnalysisParams

        # Create the output analysis result object
        result = ProximityAnalysisResult()

        # Set the start time of the analysis
        result.analysis_start = dt.datetime.utcnow()

        for traj in trajectories:
            assert type(traj) is movdata.base.Trajectory

            subject_id = traj.relocs.subject_id

            for seg in traj.traj_segs:
                for sf in proximity_analysis_params.spatial_features:
                    # diff_geo = seg.ogr_geometry.Difference(sf)
                    # if diff_geo is not None:
                    #     if diff_geo.GetGeometryName() == 'LINE':
                    #         pass

                    # Calculate the distance between the traj seg and the spatial feature
                    proximity_dist = seg.ogr_geometry.Distance(sf.ogr_geometry)

                    # Convert the distance from degrees to meters
                    proximity_dist = movdata.utils.degrees_to_km(proximity_dist) * 1000.0

                    # Create the proximity event
                    prox_event = ProximityEvent(
                        subject_id=subject_id,
                        subject_speed=seg.speed_kmhr,
                        subject_travel_heading=seg.heading,
                        proximity_distance_meters=proximity_dist,
                        proximal_fix=seg.start_fix,  # ToDo: figure out the estimated fix interpolated along the seg
                        spatial_feature_id=sf.unique_id,
                        spatial_feature_name=sf.name
                    )

                    # Add this given crossing to the result
                    result.add_proximity_event(prox_event)

        # Set the end time of the analysis
        result.analysis_end = dt.datetime.utcnow()

        return result





import os
from setuptools import setup, find_packages
import re

version_file = open("movdata/version.py").read()
version_data = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", version_file))

try:
    descr = open(os.path.join(os.path.dirname(__file__), "README.rst")).read()
except IOError:
    descr = ""

setup(
    name='movdata',
    version=version_data['version'],
    description='Movement Ecology Tools For Python',
    long_description=descr,
    url='https://github.com/wildlife-dynamics/movdata',
    author='Jake Wall',
    author_email='movementecologytools@gmail.com',
    license='MIT',
    packages=['movdata'],
    platforms=u'Posix; MacOS X; Windows',
    install_requires=['geopandas', 'gdal', 'fiona', 'geographiclib', 'numpy',
                      'scipy', 'scikit-learn', 'geojson', 'descartes', 'python-dateutil', 'rasterio', 'rasterstats',
                      'joblib', 'pyOpenSSL', 'google-api-python-client', 'earthengine-api', 'oauth2client'],
    zip_safe=False,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)



import os
import tempfile
import math
import typing
import numpy as np
import datetime as dt
from movdata import base
import shutil
from pathlib import Path
from joblib import Parallel, delayed
from joblib import load, dump
from dataclasses import dataclass, field
import movdata.raster as rst


@dataclass(init=True, repr=True)
class KDEAnalysisProfile(base.AnalysisParams):
    """Defines the outline of KDE Analysis"""
    output_path: typing.Union[str, bytes, os.PathLike]
    smooth_param: float = 0.
    smooth_method: str = 'hRef'
    max_sd: float = 5.
    prob_dens_cut_off: float = 1E-15
    raster_profile: rst.RasterProfile = None
    expansion_factor: float = 1.
    _smooth_param: float = field(init=False, repr=False)

    def __post_init__(self):
        self.raster_profile = self.raster_profile

    @property
    def smooth_param(self):
        return self._smooth_param

    @smooth_param.setter
    def smooth_param(self, value):
        self._smooth_param = value


class KDEAnalysisResult(base.AnalysisResult):

    def __init__(self, out_rast=None):
        super().__init__()
        self._out_rast = out_rast

    @property
    def out_rastdataset(self):
        return self._out_rast

    @out_rastdataset.setter
    def out_rastdataset(self, value):
        self._out_rast = value


class KDEAnalysis:

    @classmethod
    def calculate_kde_range(cls, kde_profile: KDEAnalysisProfile, relocations: base.Relocations):
        assert isinstance(kde_profile, KDEAnalysisProfile)

        folder = tempfile.mkdtemp()
        kde_result = KDEAnalysisResult()
        kde_result.analysis_start = dt.datetime.utcnow()

        try:
            # reproject relocs to desired crs
            relocs = relocations.to_crs(kde_profile.raster_profile.crs)

            # retrieve the (x,y) coordinates of relocs into numpy array
            relocs_coords = np.column_stack((relocs.geometry.x, relocs.geometry.y))

            # Determine the envelope of the relocations
            x_min, y_min, x_max, y_max = relocs.geometry.total_bounds

            # apply expansion factor on the relocation total bound.
            if kde_profile.expansion_factor > 1.0:
                dx = (x_max - x_min) * (kde_profile.expansion_factor - 1.0) / 2.0
                dy = (y_max - y_min) * (kde_profile.expansion_factor - 1.0) / 2.0
                x_min -= dx
                x_max += dx
                y_min -= dy
                y_max += dy

            # update the raster extent for the raster profile
            kde_profile.raster_profile.raster_extent = rst.RasterExtent(x_min=x_min,
                                                                        x_max=x_max,
                                                                        y_min=y_min,
                                                                        y_max=y_max)

            # determine the output raster size
            num_rows, num_columns = kde_profile.raster_profile.rows, kde_profile.raster_profile.columns

            # Define the affine transform to get grid pixel centroids as geographic coordinates and adjusting
            # the origin by a half pixel to the east and south from the grid origin point
            img_centroids_to_map = np.array(kde_profile.raster_profile.transform.to_gdal()).reshape(2, 3)
            img_centroids_to_map[0, 0] = x_min + kde_profile.raster_profile.pixel_size * 0.5
            img_centroids_to_map[1, 0] = y_max - kde_profile.raster_profile.pixel_size * 0.5

            # Calculate smoothing parameter.
            if kde_profile.smooth_param == 0.0:
                if kde_profile.smooth_method == 'hRef':
                    kde_profile.smooth_param = cls.calc_hRef_smooth_param(x=relocs_coords[:, 0], y=relocs_coords[:, 1])
                else:
                    raise TypeError(f'KDEAnalysis Error: {kde_profile.smooth_method} is not supported.')
            if kde_profile.smooth_param == 0.0:
                raise base.PymetException('KDEAnalysis Error: smoothing parameter is zero.')

            # define scaling parameter
            outer_scaling_const = 1.0 / (relocs.shape[0] *
                                         kde_profile.smooth_param *
                                         kde_profile.smooth_param)

            inner_scaling_const = 1.0 / (2.0 * np.pi)

            # define max distance value
            variance = 2.0 * kde_profile.smooth_param * kde_profile.smooth_param
            max_dist = math.pow(kde_profile.max_sd, 2.0) * variance

            # create a temporary file for storing memmap files.
            relocs_coords_tmp = os.path.join(folder, 'traj_coords.mmap')
            raster_ndarray_tmp = os.path.join(folder, 'raster_ndarray.mmap')

            # Create a blank grid array and dump the data array to memmap to speedup the parallel process.
            raster_ndarray = np.zeros(shape=(num_rows, num_columns), dtype=np.float64)
            dump(raster_ndarray, raster_ndarray_tmp)
            raster_ndarray = load(raster_ndarray_tmp, mmap_mode='w+')

            if len(relocs_coords) >= 100000:
                # Dump the trajectory coordinates to disk to free the memory
                dump(relocs_coords, relocs_coords_tmp)

                # Release the reference on the original in memory array and replace it
                # by a reference to the memmap array so that the garbage collector can
                # release the memory before forking. gc.collect() is internally called
                # in Parallel just before forking.
                relocs_coords = load(relocs_coords_tmp, mmap_mode='r')

            # Fork the worker processes to perform computation concurrently
            parameters = ((i, j, raster_ndarray, img_centroids_to_map,
                           relocs_coords,
                           max_dist,
                           inner_scaling_const,
                           outer_scaling_const,
                           variance) for i in range(num_rows)
                          for j in range(num_columns))

            Parallel(n_jobs=-1, verbose=8)(delayed(cls.raster_calc)(*p) for p in parameters)

            # Normalize the grid values
            raster_ndarray = raster_ndarray / raster_ndarray.sum()

            # write raster_ndarray to GeoTIFF file.
            rst.RasterPy.write(ndarray=raster_ndarray, fp=kde_profile.output_path, **kde_profile.raster_profile)

        except base.PymetException as exc:
            kde_result.add_error(exc)
        finally:
            shutil.rmtree(folder, ignore_errors=True)

            # Return the analysis kde_analysis_result
            kde_result.analysis_end = dt.datetime.utcnow()
            return kde_result

    @classmethod
    def calc_hRef_smooth_param(cls, x=None, y=None):
        if x is None:
            return
        if y is None:
            return
        varX = x.var()
        varY = y.var()
        var = math.sqrt(0.5 * (varX + varY))
        t = math.pow(len(x), -1.0 / 6.0)
        hRef = var * t  # Source: Worton_1989
        return hRef

    @classmethod
    def raster_calc(cls, i=0, j=0, raster_ndarray=None, img_centroids_to_map=None, traj_coords=None,
                    max_dist=None, inner_scaling_constant=None, outer_scaling_constant=None, variance=None):
        """
        Loop over each pixel in the grid, transform the pixel coordinates (row, column) into
        geographic coordinates (i.e., the centroid of the pixel). For each pixel, loop through the
        relocation points and add the KDE density contribution to the pixel value.
        """

        # # Define the grid pixel coords:  [1, Pixel (column), Line (row)]
        pixel_coords = np.array([1, j, i]).reshape(3, 1)

        # Get the current pixel centroid coordinates
        centroid = np.dot(img_centroids_to_map, pixel_coords).flatten()

        coordinate_distances = np.subtract(traj_coords[:, 0], centroid[0]) ** 2 + \
                               np.subtract(traj_coords[:, 1], centroid[1]) ** 2

        # Set flag for elements that meet out max-dist criteria.
        # ...and then eliminate the distances outside that threshold
        low_v = coordinate_distances <= max_dist
        coordinate_distances = coordinate_distances[low_v]

        # Create exponents from distance values.
        coordinate_distances = coordinate_distances * -1 / variance

        # Exp.
        coordinate_distances = np.exp(coordinate_distances)

        # Scalar sum of array.
        theta_accumulator = np.sum(coordinate_distances)

        # Set value in shared array (with scaling constant).
        raster_ndarray.itemset((i, j), raster_ndarray[i, j] + theta_accumulator * outer_scaling_constant)



import movdata as pm

class GPSTrackingDataQualityAnalyzer:
    """Methods for testing the quality of tracking data """

    def __init__(self,trajectory):
        traj = trajectory

    def CheckForDataGaps(self):
        """Function to examine a trajectory for gaps in the data"""
        pass

    def ScheduleAdherence(self, SampleSchedule):
        #Function to compare
        pass

